"""Module providing adapter class making edge-label prediction possible in sklearn models."""
from sklearn.base import ClassifierMixin
from typing import Type, Union, List
from embiggen.utils.sklearn_utils import must_be_an_sklearn_classifier_model
from embiggen.edge_label_prediction.sklearn_like_edge_label_prediction_adapter import (
    SklearnLikeEdgeLabelPredictionAdapter,
)


class SklearnEdgeLabelPredictionAdapter(SklearnLikeEdgeLabelPredictionAdapter):
    """Class wrapping Sklearn models for running edge-label prediction."""

    def __init__(
        self,
        model_instance: Type[ClassifierMixin],
        edge_embedding_methods: Union[List[str], str] = "Concatenate",
        use_edge_metrics: bool = False,
        random_state: int = 42,
    ):
        """Create the adapter for Sklearn object.

        Parameters
        ----------------
        model_instance: Type[ClassifierMixin]
            The class instance to be adapted into edge-label prediction.
        edge_embedding_method: Union[List[str], str] = "Concatenate",
            The method(s) to use to compute the edges.
            If multiple edge embedding are provided, they
            will be Concatenated and fed to the model.
            The supported edge embedding methods are:
             * Hadamard: element-wise product
             * Sum: element-wise sum
             * Average: element-wise mean
             * L1: element-wise subtraction
             * AbsoluteL1: element-wise subtraction in absolute value
             * SquaredL2: element-wise subtraction in squared value
             * L2: element-wise squared root of squared subtraction
             * Concatenate: Concatenate of source and destination node features
             * Min: element-wise minimum
             * Max: element-wise maximum
             * L2Distance: vector-wise L2 distance - this yields a scalar
             * CosineSimilarity: vector-wise cosine similarity - this yields a scalar
        use_edge_metrics: bool = False
            Whether to use the edge metrics from traditional edge prediction.
            These metrics currently include:
            - Adamic Adar
            - Jaccard Coefficient
            - Resource allocation index
            - Preferential attachment
        random_state: int = 42
            The random state to use to reproduce the training.

        Raises
        ----------------
        ValueError
            If the provided model_instance is not a subclass of `ClassifierMixin`.
        """
        must_be_an_sklearn_classifier_model(model_instance)
        # We want to mask the decorator class name
        self.__class__.__name__ = model_instance.__class__.__name__
        self.__class__.__doc__ = model_instance.__class__.__doc__
        super().__init__(
            model_instance=model_instance,
            random_state=random_state,
            edge_embedding_methods=edge_embedding_methods,
            use_edge_metrics=use_edge_metrics,
        )

    @classmethod
    def library_name(cls) -> str:
        """Return name of the model."""
        return "scikit-learn"


"""Edge prediction model based on CatBoost."""
from typing import Dict, Any, Optional, List, Union
from catboost import CatBoostClassifier
from embiggen.edge_prediction.sklearn_like_edge_prediction_adapter import (
    SklearnLikeEdgePredictionAdapter,
)


class CatBoostEdgePrediction(SklearnLikeEdgePredictionAdapter):
    """Edge prediction model based on CatBoost."""

    def __init__(
        self,
        iterations: int = 500,
        learning_rate: float = 0.03,
        max_depth: int = 6,
        l2_leaf_reg: float = 3.0,
        model_size_reg: float = 0.5,
        rsm: float = 1.0,
        loss_function: Optional[str] = None,
        border_count: int = 254,
        feature_border_type: str = "GreedyLogSum",
        per_float_feature_quantization: Optional[List[str]] = None,
        input_borders: Optional[str] = None,
        output_borders: Optional[str] = None,
        fold_permutation_block: int = 1,
        od_pval: Optional[float] = None,
        od_wait: Optional[int] = None,
        od_type: Optional[str] = None,
        nan_mode: str = "Min",
        counter_calc_method: str = "SkipTest",
        leaf_estimation_iterations: int = 1,
        leaf_estimation_method: str = "Newton",
        thread_count: int = -1,
        use_best_model: Optional[bool] = None,
        best_model_min_trees: int = 1,
        verbose: bool = False,
        metric_period: int = 1,
        ctr_leaf_count_limit: int = 16,
        store_all_simple_ctr: bool = False,
        max_ctr_complexity: int = 4,
        has_time: bool = False,
        allow_const_label: bool = False,
        target_border=None,
        classes_count: Optional[int] = None,
        class_weights=None,
        auto_class_weights: str = "Balanced",
        class_names=None,
        one_hot_max_size: Optional[int] = None,
        random_strength: float = 1.0,
        name: str = "experiment",
        ignored_features=None,
        train_dir: Optional[str] = None,
        custom_metric=None,
        eval_metric=None,
        bagging_temperature: int = 1,
        save_snapshot=None,
        snapshot_file=None,
        snapshot_interval: int = 600,
        fold_len_multiplier=None,
        used_ram_limit=None,
        gpu_ram_part: float = 0.95,
        pinned_memory_size=None,
        allow_writing_files=None,
        final_ctr_computation_mode: str = "Default",
        approx_on_full_history: bool = False,
        boosting_type=None,
        simple_ctr=None,
        combinations_ctr: Optional[List[str]] = None,
        per_feature_ctr: Optional[List[str]] = None,
        ctr_target_border_count: int = 1,
        task_type: Optional[str] = None,
        devices=None,
        bootstrap_type: str = "MVS",
        subsample: float = 1.0,
        mvs_reg=None,
        sampling_unit: str = "Object",
        sampling_frequency: str = "PerTree",
        dev_score_calc_obj_block_size: int = 5000000,
        dev_efb_max_buckets: int = 1024,
        sparse_features_conflict_fraction: float = 0.0,
        early_stopping_rounds=None,
        cat_features=None,
        grow_policy: str = "SymmetricTree",
        min_data_in_leaf: int = 1,
        max_leaves: int = 31,
        score_function: str = "Cosine",
        leaf_estimation_backtracking=None,
        monotone_constraints=None,
        feature_weights=None,
        penalties_coefficient: float = 1.0,
        first_feature_use_penalties=None,
        per_object_feature_penalties=None,
        model_shrink_rate: float = 0.0,
        model_shrink_mode=None,
        langevin=None,
        diffusion_temperature: float = 0.0,
        posterior_sampling=None,
        boost_from_average=None,
        text_features=None,
        tokenizers=None,
        dictionaries=None,
        feature_calcers=None,
        text_processing=None,
        embedding_features=None,
        callback=None,
        eval_fraction=None,
        edge_embedding_methods: Union[List[str], str] = "Concatenate",
        training_unbalance_rate: float = 1.0,
        training_sample_only_edges_with_heterogeneous_node_types: bool = False,
        use_edge_metrics: bool = False,
        use_scale_free_distribution: bool = True,
        prediction_batch_size: int = 2**12,
        random_state: int = 42,
    ):
        """Build a CatBoost Edge prediction model."""
        self._kwargs = dict(
            iterations=iterations,
            learning_rate=learning_rate,
            max_depth=max_depth,
            l2_leaf_reg=l2_leaf_reg,
            model_size_reg=model_size_reg,
            rsm=rsm,
            loss_function=loss_function,
            border_count=border_count,
            feature_border_type=feature_border_type,
            per_float_feature_quantization=per_float_feature_quantization,
            input_borders=input_borders,
            output_borders=output_borders,
            fold_permutation_block=fold_permutation_block,
            od_pval=od_pval,
            od_wait=od_wait,
            od_type=od_type,
            nan_mode=nan_mode,
            counter_calc_method=counter_calc_method,
            leaf_estimation_iterations=leaf_estimation_iterations,
            leaf_estimation_method=leaf_estimation_method,
            thread_count=thread_count,
            use_best_model=use_best_model,
            best_model_min_trees=best_model_min_trees,
            verbose=verbose,
            metric_period=metric_period,
            ctr_leaf_count_limit=ctr_leaf_count_limit,
            store_all_simple_ctr=store_all_simple_ctr,
            max_ctr_complexity=max_ctr_complexity,
            has_time=has_time,
            allow_const_label=allow_const_label,
            target_border=target_border,
            classes_count=classes_count,
            class_weights=class_weights,
            auto_class_weights=auto_class_weights,
            class_names=class_names,
            one_hot_max_size=one_hot_max_size,
            random_strength=random_strength,
            name=name,
            ignored_features=ignored_features,
            train_dir=train_dir,
            custom_metric=custom_metric,
            eval_metric=eval_metric,
            bagging_temperature=bagging_temperature,
            save_snapshot=save_snapshot,
            snapshot_file=snapshot_file,
            snapshot_interval=snapshot_interval,
            fold_len_multiplier=fold_len_multiplier,
            used_ram_limit=used_ram_limit,
            gpu_ram_part=gpu_ram_part,
            pinned_memory_size=pinned_memory_size,
            allow_writing_files=allow_writing_files,
            final_ctr_computation_mode=final_ctr_computation_mode,
            approx_on_full_history=approx_on_full_history,
            boosting_type=boosting_type,
            simple_ctr=simple_ctr,
            combinations_ctr=combinations_ctr,
            per_feature_ctr=per_feature_ctr,
            ctr_target_border_count=ctr_target_border_count,
            task_type=task_type,
            devices=devices,
            bootstrap_type=bootstrap_type,
            subsample=subsample,
            mvs_reg=mvs_reg,
            sampling_unit=sampling_unit,
            sampling_frequency=sampling_frequency,
            dev_score_calc_obj_block_size=dev_score_calc_obj_block_size,
            dev_efb_max_buckets=dev_efb_max_buckets,
            sparse_features_conflict_fraction=sparse_features_conflict_fraction,
            early_stopping_rounds=early_stopping_rounds,
            cat_features=cat_features,
            grow_policy=grow_policy,
            min_data_in_leaf=min_data_in_leaf,
            max_leaves=max_leaves,
            score_function=score_function,
            leaf_estimation_backtracking=leaf_estimation_backtracking,
            monotone_constraints=monotone_constraints,
            feature_weights=feature_weights,
            penalties_coefficient=penalties_coefficient,
            first_feature_use_penalties=first_feature_use_penalties,
            per_object_feature_penalties=per_object_feature_penalties,
            model_shrink_rate=model_shrink_rate,
            model_shrink_mode=model_shrink_mode,
            langevin=langevin,
            diffusion_temperature=diffusion_temperature,
            posterior_sampling=posterior_sampling,
            boost_from_average=boost_from_average,
            text_features=text_features,
            tokenizers=tokenizers,
            dictionaries=dictionaries,
            feature_calcers=feature_calcers,
            text_processing=text_processing,
            embedding_features=embedding_features,
            callback=callback,
            eval_fraction=eval_fraction,
        )

        super().__init__(
            model_instance=CatBoostClassifier(
                **self._kwargs,
                random_state=random_state,
            ),
            edge_embedding_methods=edge_embedding_methods,
            training_unbalance_rate=training_unbalance_rate,
            training_sample_only_edges_with_heterogeneous_node_types=training_sample_only_edges_with_heterogeneous_node_types,
            use_edge_metrics=use_edge_metrics,
            use_scale_free_distribution=use_scale_free_distribution,
            prediction_batch_size=prediction_batch_size,
        )

    @classmethod
    def smoke_test_parameters(cls) -> Dict[str, Any]:
        return dict(
            iterations=2,
            max_depth=2,
        )

    def parameters(self) -> Dict[str, Any]:
        return dict(
            **super().parameters(),
            **self._kwargs,
        )

    @classmethod
    def model_name(cls) -> str:
        """Return the name of the model."""
        return "CatBoost"

    @classmethod
    def library_name(cls) -> str:
        """Return name of the model."""
        return "CatBoost"


"""Submodule wrapping Radius Neighbour for Edge prediction."""
from typing import Dict, Any, Optional, Union, List
from sklearn.neighbors import RadiusNeighborsClassifier
from embiggen.edge_label_prediction.edge_label_prediction_sklearn.sklearn_edge_label_prediction_adapter import (
    SklearnEdgeLabelPredictionAdapter,
)
from embiggen.utils import normalize_kwargs


class RadiusNeighborsClassifierEdgePrediction(SklearnEdgeLabelPredictionAdapter):
    """Create wrapper over Sklearn Radius Neighbour classifier for Edge prediction."""

    def __init__(
        self,
        radius: float = 1.0,
        weights: str = "uniform",
        algorithm: str = "auto",
        leaf_size: int = 30,
        p: int = 2,
        metric: str = "minkowski",
        outlier_label: Optional[str] = "most_frequent",
        metric_params: Dict[str, Any] = None,
        n_jobs: int = -1,
        edge_embedding_methods: Union[List[str], str] = "Concatenate",
        use_edge_metrics: bool = False,
        random_state: int = 42,
    ):
        """Create the Decision Tree for Edge prediction."""
        self._kwargs = normalize_kwargs(
            self,
            dict(
                radius=radius,
                weights=weights,
                algorithm=algorithm,
                leaf_size=leaf_size,
                p=p,
                metric=metric,
                outlier_label=outlier_label,
                metric_params=metric_params,
                n_jobs=n_jobs,
            )
        )

        super().__init__(
            RadiusNeighborsClassifier(**self._kwargs),
            edge_embedding_methods=edge_embedding_methods,
            use_edge_metrics=use_edge_metrics,
            random_state=random_state,
        )

    def parameters(self) -> Dict[str, Any]:
        """Returns parameters used for this model."""
        return {**super().parameters(), **self._kwargs}

    @classmethod
    def smoke_test_parameters(cls) -> Dict[str, Any]:
        """Returns parameters for smoke test."""
        return dict(leaf_size=1000)

    @classmethod
    def model_name(cls) -> str:
        return "Radius Neighbour Classifier"



"""Submodule to test whether the metaprogramming is working properly."""
from embiggen.utils.abstract_models import build_init, AbstractEmbeddingModel

build_init(
    module_library_names="non_existent_module",
    formatted_library_name="Non Existent Module",
    task_name="Node Embedding",
    expected_parent_class=AbstractEmbeddingModel,
)



"""NodeTransformer class to convert nodes to edge embeddings."""
from typing import List, Union, Optional
import numpy as np
import pandas as pd
from ensmallen import Graph


class NodeTransformer:
    """NodeTransformer class to convert nodes to edge embeddings."""

    def __init__(
        self,
        aligned_mapping: bool = False,
    ):
        """Create new NodeTransformer object.

        Parameters
        -------------------
        aligned_mapping: bool = False,
            This parameter specifies whether the mapping of the embeddings nodes
            matches the internal node mapping of the given graph.
            If these two mappings do not match, the generated edge embedding
            will be meaningless.
        """
        self._node_feature = []
        self._node_type_feature = []
        self._aligned_mapping = aligned_mapping

    def fit(
        self,
        node_feature: Optional[Union[pd.DataFrame, np.ndarray, List[Union[pd.DataFrame, np.ndarray]]]] = None,
        node_type_feature: Optional[Union[pd.DataFrame, np.ndarray, List[Union[pd.DataFrame, np.ndarray]]]] = None,
    ):
        """Fit the model.

        Parameters
        -------------------------
        node_feature: Optional[Union[pd.DataFrame, np.ndarray, List[Union[pd.DataFrame, np.ndarray]]]] = None
            Node feature to use to fit the transformer.
        node_type_feature: Optional[Union[pd.DataFrame, np.ndarray, List[Union[pd.DataFrame, np.ndarray]]]] = None
            Node type feature to use to fit the transformer.
        """
        if node_feature is None:
            node_feature = []
        if not isinstance(node_feature, list):
            node_feature = [node_feature]

        if node_type_feature is None:
            node_type_feature = []
        if not isinstance(node_type_feature, list):
            node_type_feature = [node_type_feature]

        # We check if any of the provided node features
        # is neither a numpy array nor a pandas dataframe.
        for features, feature_name in (
            (node_type_feature, "node type"),
            (node_feature, "node"),
        ):
            for feature in features:
                if not isinstance(feature, (pd.DataFrame, np.ndarray)):
                    raise ValueError(
                        (
                            f"One of the provided {feature_name} features is not "
                            "neither a pandas DataFrame nor a numpy array, but "
                            f"of type {type(feature)}. It is not clear "
                            "what to do with this feature."
                        )
                    )
                # We check whether the provided features contain any NaN
                if isinstance(feature, pd.DataFrame) and feature.isna().any().any():
                    raise ValueError(
                        (
                            f"One of the provided {feature_name} features contains NaNs. "
                            "This is not supported. The DataFrame has shape "
                            f"{feature.shape} and the first 5 rows are:\n"
                            f"{feature.head(5)}"
                        )
                    )
                
                if isinstance(feature, np.ndarray) and np.isnan(feature).any():
                    raise ValueError(
                        (
                            f"One of the provided {feature_name} features contains NaNs. "
                            "This is not supported. The array has shape "
                            f"{feature.shape} and the first 5 rows are:\n"
                            f"{feature[:5]}"
                        )
                    )
                

            # We check if, while the parameters for alignment
            # has not been provided, numpy arrays were provided.
            # This would be an issue as we cannot check for alignment
            # in numpy arrays.
            if not self._aligned_mapping and any(
                isinstance(nf, np.ndarray)
                for nf in features
            ):
                raise ValueError(
                    "A numpy array feature was provided while the "
                    f"aligned mapping parameter was set to false. "
                    "If you intend to specify that you are providing a numpy "
                    f"array {feature_name} feature that is aligned with the vocabulary "
                    "of the graph set the `aligned_mapping` parameter "
                    "to True."
                )

        if self._aligned_mapping:
            if len(node_feature) > 1:
                self._node_feature = np.hstack([
                    nf.to_numpy() if isinstance(nf, pd.DataFrame) else nf
                    for nf in node_feature
                ])
            elif len(node_feature) == 1:
                self._node_feature = node_feature[0]

            if len(node_type_feature) > 1:
                self._node_type_feature = np.hstack([
                    nf.to_numpy() if isinstance(nf, pd.DataFrame) else nf
                    for nf in node_type_feature
                ])
            elif len(node_type_feature) == 1:
                self._node_type_feature = node_type_feature[0]
        else:
            if len(node_feature) > 0:
                self._node_feature = pd.concat(node_feature, axis=1)
            if len(node_type_feature) > 0:
                self._node_type_feature = pd.concat(node_type_feature, axis=1)

    def has_node_type_features(self) -> bool:
        """Return whether the transformer has node type feature."""
        return len(self._node_type_feature) > 0
    
    def has_node_features(self) -> bool:
        """Return whether the transformer has node feature."""
        return len(self._node_feature) > 0

    def is_aligned_mapping(self) -> bool:
        """Return whether the transformer can assume aligned mapping."""
        return self._aligned_mapping

    def is_fit(self) -> bool:
        """Return whether the transformer is fitted."""
        return len(self._node_feature) + len(self._node_type_feature) > 0

    def transform(
        self,
        nodes: Optional[Union[Graph, List[str], List[int]]] = None,
        node_types: Optional[Union[Graph, List[Optional[List[str]]], List[Optional[List[int]]]]] = None,
    ) -> np.ndarray:
        """Return embeddings from given node.

        Parameters
        --------------------------
        nodes: Optional[Union[List[str], List[int]]] = None
            List of nodes whose embedding is to be returned.
            This can be either a list of strings, or a graph, or if the
            aligned_mapping is setted, then this methods also accepts
            a list of ints.
        node_types: Optional[Union[Graph, List[Optional[List[str]]], List[Optional[List[int]]]]] = None,
            List of node types whose embedding is to be returned.
            This can be either a list of strings, or a graph, or if the
            aligned_mapping is setted, then this methods also accepts
            a list of ints.

        Raises
        --------------------------
        ValueError,
            If embedding is not fitted.

        Returns
        --------------------------
        Numpy array of embeddings.
        """
        if not self.is_fit():
            raise ValueError(
                "Transformer was not fitted yet."
            )

        node_type_features = None
        node_features = None

        if self._aligned_mapping:
            if nodes is not None and self.has_node_features():
                if not isinstance(nodes, (np.ndarray, Graph)):
                    raise ValueError(
                        "The provided nodes are not numpy array and the "
                        "node IDs or Graph are expected to be aligned."
                    )

                if isinstance(nodes, Graph):
                    node_features = self._node_feature
                else:
                    node_features = self._node_feature[nodes]

            if node_types is not None and self.has_node_type_features():
                if isinstance(node_types, Graph):
                    node_types = node_types.get_node_type_ids()
                    if isinstance(nodes, np.ndarray):
                        node_types = [
                            node_types[node]
                            for node in nodes
                        ]
                node_type_feature_dimensionality = self._node_type_feature.shape[1]
                node_type_features = np.vstack([
                    np.mean(
                        self._node_type_feature[node_type_ids],
                        axis=0
                    )
                    if node_type_ids is not None
                    else
                    np.zeros(shape=node_type_feature_dimensionality)
                    for node_type_ids in node_types
                ])
        else:
            if nodes is not None and self.has_node_features():
                if isinstance(nodes, Graph):
                    nodes = nodes.get_node_names()

                node_features = self._node_feature.loc[nodes].to_numpy()

            if node_types is not None and self.has_node_type_features():
                node_type_feature_dimensionality = self._node_type_feature.shape[1]
                node_type_features = np.vstack([
                    np.mean(
                        self._node_feature.loc[node_type_names].to_numpy(),
                        axis=0
                    )
                    if node_type_names is not None
                    else
                    np.zeros(shape=node_type_feature_dimensionality)
                    for node_type_names in node_types
                ])

        if node_features is None:
            node_features = node_type_features
        elif node_type_features is not None:
            node_features = np.hstack([
                node_features,
                node_type_features
            ])

        return node_features



"""EdgePredictionTransformer class to convert graphs to edge embeddings to execute edge prediction."""
from typing import Tuple, Union, List, Optional
import pandas as pd
import numpy as np
from ensmallen import Graph  # pylint: disable=no-name-in-module

from embiggen.embedding_transformers.graph_transformer import GraphTransformer


class EdgePredictionTransformer:
    """EdgePredictionTransformer class to convert graphs to edge embeddings."""

    def __init__(
        self,
        methods: Union[List[str], str] = "Hadamard",
        aligned_mapping: bool = False,
        include_both_undirected_edges: bool = True
    ):
        """Create new EdgePredictionTransformer object.

        Parameters
        ------------------------
        methods: Union[List[str], str] = "Hadamard",
            Method to use for the embedding.
            If None is used, we return instead the numeric tuples.
            If multiple edge embedding are provided, they
            will be Concatenated and fed to the model.
            The supported edge embedding methods are:
             * Hadamard: element-wise product
             * Sum: element-wise sum
             * Average: element-wise mean
             * L1: element-wise subtraction
             * AbsoluteL1: element-wise subtraction in absolute value
             * SquaredL2: element-wise subtraction in squared value
             * L2: element-wise squared root of squared subtraction
             * Concatenate: Concatenate of source and destination node features
             * Min: element-wise minimum
             * Max: element-wise maximum
             * L2Distance: vector-wise L2 distance - this yields a scalar
             * CosineSimilarity: vector-wise cosine similarity - this yields a scalar
        aligned_mapping: bool = False,
            This parameter specifies whether the mapping of the embeddings nodes
            matches the internal node mapping of the given graph.
            If these two mappings do not match, the generated edge embedding
            will be meaningless.
        include_both_undirected_edges: bool = True
            Whether to include both directed and undirected edges.
        """
        self._transformer = GraphTransformer(
            methods=methods,
            aligned_mapping=aligned_mapping,
            include_both_undirected_edges=include_both_undirected_edges
        )

    def fit(
        self,
        node_feature: Union[pd.DataFrame, np.ndarray, List[Union[pd.DataFrame, np.ndarray]]],
        node_type_feature: Optional[Union[pd.DataFrame, np.ndarray,
                                          List[Union[pd.DataFrame, np.ndarray]]]] = None,
        edge_type_features: Optional[Union[pd.DataFrame, np.ndarray,
                                           List[Union[pd.DataFrame, np.ndarray]]]] = None,
    ):
        """Fit the model.

        Parameters
        -------------------------
        node_feature: Union[pd.DataFrame, np.ndarray, List[Union[pd.DataFrame, np.ndarray]]],
            Node feature to use to fit the transformer.
        node_type_feature: Optional[Union[pd.DataFrame, np.ndarray, List[Union[pd.DataFrame, np.ndarray]]]] = None
            Node type feature to use to fit the transformer.
        edge_type_features: Optional[Union[pd.DataFrame, np.ndarray, List[Union[pd.DataFrame, np.ndarray]]]] = None
            Edge type feature to use to fit the transformer.

        Raises
        -------------------------
        ValueError
            If the given method is None there is no need to call the fit method.
        """
        self._transformer.fit(
            node_feature,
            node_type_feature=node_type_feature,
            edge_type_features=edge_type_features
        )

    def transform(
        self,
        positive_graph: Union[Graph, np.ndarray, List[List[str]], List[List[int]]],
        negative_graph: Union[Graph, np.ndarray, List[List[str]], List[List[int]]],
        edge_features: Optional[Union[np.ndarray, List[np.ndarray]]] = None,
        random_state: int = 42,
        shuffle: bool = False
    ) -> Tuple[np.ndarray, np.ndarray]:
        """Return edge embedding for given graph using provided method.

        Parameters
        --------------------------
        positive_graph: Union[Graph, List[List[str]], List[List[int]]],
            The graph whose edges are to be embedded and labeled as positives.
            It can either be an Graph or a list of lists of edges.
        negative_graph: Union[Graph, List[List[str]], List[List[int]]],
            The graph whose edges are to be embedded and labeled as positives.
            It can either be an Graph or a list of lists of edges.
        edge_features: Optional[Union[np.ndarray, List[np.ndarray]]] = None
            Optional edge features to be used as input Concatenated
            to the obtained edge embedding. The shape must be equal
            to the number of directed edges in the provided graph.
        random_state: int = 42,
            The random state to use to shuffle the labels.
        shuffle: bool = False
            Whether to shuffle the samples

        Raises
        --------------------------
        ValueError,
            If embedding is not fitted.

        Returns
        --------------------------
        Tuple with X and y values.
        """
        if isinstance(positive_graph, Graph) and isinstance(negative_graph, Graph):
            if not positive_graph.has_compatible_node_vocabularies(negative_graph):
                raise ValueError(
                    "The provided positive and negative graphs are not compatible. "
                    "Possible causes for this may be a different node vocabulary, "
                    "presence or absence of node types or different node type vocabulary, "
                    "the presence or absence of edge weights or types in one of the "
                    "two graphs."
                )

        if edge_features is None:
            edge_features = []

        if not isinstance(edge_features, list):
            edge_features = [edge_features]

        if isinstance(positive_graph, Graph):
            number_of_positive_edges = positive_graph.get_number_of_edges()
        else:
            number_of_positive_edges = len(positive_graph)

        if isinstance(negative_graph, Graph):
            number_of_negative_edges = negative_graph.get_number_of_edges()
        else:
            number_of_negative_edges = len(negative_graph)

        for edge_feature in edge_features:
            if not isinstance(edge_feature, np.ndarray):
                raise ValueError(
                    "The provided edge features must be a numpy array. "
                    f"Your provided edge features are of type {type(edge_feature)}."
                )

            if edge_feature.shape[0] != number_of_positive_edges + number_of_negative_edges:
                raise ValueError(
                    "The provided edge features must have the same number of edges as the provided graphs. "
                    f"The provided edge features have {edge_feature.shape[0]} edges, while the provided graphs have "
                    f"{number_of_positive_edges} and {number_of_negative_edges} edges."
                )
            
        positive_edge_features = [
            edge_feature[:number_of_positive_edges]
            for edge_feature in edge_features
        ]

        negative_edge_features = [
            edge_feature[number_of_positive_edges:]
            for edge_feature in edge_features
        ]

        positive_edge_embedding = self._transformer.transform(
            positive_graph,
            node_types=positive_graph if self._transformer.has_node_type_features() else None,
            edge_types=positive_graph if self._transformer.has_edge_type_features() else None,
            edge_features=positive_edge_features
        )
        negative_edge_embedding = self._transformer.transform(
            negative_graph,
            node_types=negative_graph if self._transformer.has_node_type_features() else None,
            edge_types=negative_graph if self._transformer.has_edge_type_features() else None,
            edge_features=negative_edge_features
        )

        edge_embeddings = np.vstack([
            positive_edge_embedding,
            negative_edge_embedding
        ])

        edge_labels = np.concatenate([
            np.ones(positive_edge_embedding.shape[0]),
            np.zeros(negative_edge_embedding.shape[0])
        ])

        if shuffle:
            numpy_random_state = np.random.RandomState(  # pylint: disable=no-member
                seed=random_state
            )

            indices = numpy_random_state.permutation(edge_labels.shape[0])

            edge_embeddings, edge_labels = edge_embeddings[indices], edge_labels[indices]

        return edge_embeddings, edge_labels



"""GCN model for edge prediction."""
from typing import List, Union, Optional, Dict, Any, Type

import numpy as np
from tensorflow.keras.optimizers import (
    Optimizer,
)  # pylint: disable=import-error,no-name-in-module

from ensmallen import Graph
from embiggen.edge_prediction.edge_prediction_model import AbstractEdgePredictionModel
from embiggen.utils.abstract_edge_gcn import AbstractEdgeGCN
from embiggen.sequences.tensorflow_sequences import (
    GCNEdgePredictionTrainingSequence,
    GCNEdgePredictionSequence,
)
from embiggen.utils import AbstractEdgeFeature


class GCNEdgePrediction(AbstractEdgeGCN, AbstractEdgePredictionModel):
    """GCN model for edge prediction."""

    def __init__(
        self,
        kernels: Optional[Union[str, List[str]]] = ["Symmetric Normalized Laplacian",],
        epochs: int = 1000,
        number_of_graph_convolution_layers: int = 2,
        number_of_units_per_graph_convolution_layers: Union[int, List[int]] = 128,
        number_of_ffnn_body_layers: int = 2,
        number_of_ffnn_head_layers: int = 1,
        number_of_units_per_ffnn_body_layer: Union[int, List[int]] = 128,
        number_of_units_per_ffnn_head_layer: Union[int, List[int]] = 128,
        dropout_rate: float = 0.2,
        batch_size: Optional[int] = None,
        number_of_batches_per_epoch: Optional[int] = None,
        apply_norm: bool = False,
        combiner: str = "sum",
        edge_embedding_methods: Union[List[str], str] = "Concatenate",
        optimizer: Union[str, Optimizer] = "adam",
        early_stopping_min_delta: float = 0.0001,
        early_stopping_patience: int = 20,
        reduce_lr_min_delta: float = 0.0001,
        reduce_lr_patience: int = 5,
        early_stopping_monitor: str = "loss",
        early_stopping_mode: str = "min",
        reduce_lr_monitor: str = "loss",
        reduce_lr_mode: str = "min",
        reduce_lr_factor: float = 0.9,
        avoid_false_negatives: bool = True,
        training_unbalance_rate: float = 1.0,
        training_sample_only_edges_with_heterogeneous_node_types: bool = False,
        use_edge_metrics: bool = False,
        random_state: int = 42,
        use_node_embedding: bool = False,
        node_embedding_size: int = 50,
        use_node_type_embedding: bool = False,
        node_type_embedding_size: int = 50,
        use_edge_type_embedding: bool = False,
        edge_type_embedding_size: int = 50,
        residual_convolutional_layers: bool = False,
        siamese_node_feature_module: bool = True,
        handling_multi_graph: str = "warn",
        node_feature_names: Optional[Union[str, List[str]]] = None,
        node_type_feature_names: Optional[Union[str, List[str]]] = None,
        edge_type_feature_names: Optional[Union[str, List[str]]] = None,
        verbose: bool = False,
    ):
        """Create new Kipf GCN object.

        Parameters
        -------------------------------
        kernels: Optional[Union[str, List[str]]]
            The type of normalization to use. It can either be:
            * "Weights", to just use the graph weights themselves.
            * "Left Normalized Laplacian", for the left normalized Laplacian.
            * "Right Normalized Laplacian", for the right normalized Laplacian.
            * "Symmetric Normalized Laplacian", for the symmetric normalized Laplacian.
            * "Transposed Left Normalized Laplacian", for the transposed left normalized Laplacian.
            * "Transposed Right Normalized Laplacian", for the transposed right normalized Laplacian.
            * "Transposed Symmetric Normalized Laplacian", for the transposed symmetric normalized Laplacian.
            * "Weighted Left Normalized Laplacian", for the weighted left normalized Laplacian.
            * "Weighted Right Normalized Laplacian", for the weighted right normalized Laplacian.
            * "Weighted Symmetric Normalized Laplacian", for the weighted symmetric normalized Laplacian.
            * "Transposed Weighted Left Normalized Laplacian", for the transposed weighted left normalized Laplacian.
            * "Transposed Weighted Right Normalized Laplacian", for the transposed weighted right normalized Laplacian.
            * "Transposed Weighted Symmetric Normalized Laplacian", for the transposed weighted symmetric normalized Laplacian.
        epochs: int = 1000
            Epochs to train the model for.
        number_of_graph_convolution_layers: int = 2
            Number of layers in the body subsection of the GCN section of the model.
        number_of_gcn_head_layers: int = 1
            Number of layers in the head subsection of the GCN section of the model.
        number_of_ffnn_body_layers: int = 2
            Number of layers in the body subsection of the FFNN section of the model.
        number_of_ffnn_head_layers: int = 1
            Number of layers in the head subsection of the FFNN section of the model.
        number_of_units_per_gcn_body_layer: Union[int, List[int]] = 128
            Number of units per gcn body layer.
        number_of_units_per_gcn_head_layer: Union[int, List[int]] = 128
            Number of units per gcn head layer.
        number_of_units_per_ffnn_body_layer: Union[int, List[int]] = 128
            Number of units per ffnn body layer.
        number_of_units_per_ffnn_head_layer: Union[int, List[int]] = 128
            Number of units per ffnn head layer.
        dropout_rate: float = 0.3
            Float between 0 and 1.
            Fraction of the input units to dropout.
        batch_size: Optional[int] = None
            Batch size to use while training the model.
            If None, the batch size will be the number of nodes.
            In all model parametrization that involve a number of graph
            convolution layers, the batch size will be the number of nodes.
        number_of_batches_per_epoch: Optional[int] = None
            Number of batches to use per epoch.
            By default, this is None, which means that the number of batches
            will be equal to the number of directed edges divided by the batch size.
        apply_norm: bool = False
            Whether to normalize the output of the convolution operations,
            after applying the level activations.
        combiner: str = "mean"
            A string specifying the reduction op.
            Currently "mean", "sqrtn" and "sum" are supported.
            "sum" computes the weighted sum of the embedding results for each row.
            "mean" is the weighted sum divided by the total weight.
            "sqrtn" is the weighted sum divided by the square root of the sum of the squares of the weights.
            Defaults to mean.
        edge_embedding_method: str = "Concatenate"
            The edge embedding method to use to put togheter the
            source and destination node features, which includes:
            - Concatenate
            - Average
            - Hadamard
            - L1
            - L2
            - Maximum
            - Minimum
            - Add
            - Subtract
            - Dot
        optimizer: str = "adam"
            The optimizer to use while training the model.
        early_stopping_min_delta: float
            Minimum delta of metric to stop the training.
        early_stopping_patience: int
            Number of epochs to wait for when the given minimum delta is not
            achieved after which trigger early stopping.
        reduce_lr_min_delta: float
            Minimum delta of metric to reduce learning rate.
        reduce_lr_patience: int
            Number of epochs to wait for when the given minimum delta is not
            achieved after which reducing learning rate.
        early_stopping_monitor: str = "loss",
            Metric to monitor for early stopping.
        early_stopping_mode: str = "min",
            Direction of the variation of the monitored metric for early stopping.
        reduce_lr_monitor: str = "loss",
            Metric to monitor for reducing learning rate.
        reduce_lr_mode: str = "min",
            Direction of the variation of the monitored metric for learning rate.
        reduce_lr_factor: float = 0.9,
            Factor for reduction of learning rate.
        avoid_false_negatives: bool = True
            Whether to avoid sampling false negatives.
            This check makes the sampling a bit slower, and generally
            the rate of collision is extremely low.
            Consider disabling this when the task can account for this.
        training_unbalance_rate: float = 1.0
            The amount of negatives to be sampled during the training of the model.
            By default this is 1.0, which means that the same number of positives and
            negatives in the training are of the same cardinality.
        training_sample_only_edges_with_heterogeneous_node_types: bool = False
            Whether to sample exclusively edges between nodes with different node types.
        use_node_embedding: bool = False
            Whether to use a node embedding layer to let the model automatically
            learn an embedding of the nodes.
        node_embedding_size: int = 50
            Size of the node embedding.
        use_node_types: Union[bool, str] = "auto"
            Whether to use the node types while training the model.
            By default, automatically uses them if the graph has them.
        node_type_embedding_size: int = 50
            Size of the embedding for the node types.
        use_edge_metrics: bool = False
            Whether to use the edge metrics from traditional edge prediction.
            These metrics currently include:
            - Adamic Adar
            - Jaccard Coefficient
            - Resource allocation index
            - Preferential attachment
        use_node_embedding: bool = False
            Whether to use a node embedding layer that is automatically learned
            by the model while it trains. Please do be advised that by using
            a node embedding layer you are making a closed-world assumption,
            and this model will not work on graphs with a different node vocabulary.
        node_embedding_size: int = 50
            Dimension of the node embedding.
        use_node_type_embedding: bool = False
            Whether to use a node type embedding layer that is automatically learned
            by the model while it trains. Please do be advised that by using
            a node type embedding layer you are making a closed-world assumption,
            and this model will not work on graphs with a different node vocabulary.
        node_type_embedding_size: int = 50
            Dimension of the node type embedding.
        use_edge_type_embedding: bool = False
            Whether to use a edge type embedding layer that is automatically learned
            by the model while it trains. Please do be advised that by using
            a edge type embedding layer you are making a closed-world assumption,
            and this model will not work on graphs with a different edge vocabulary.
        edge_type_embedding_size: int = 50
            Dimension of the edge type embedding.
        residual_convolutional_layers: bool = False
            Whether to use residual connections in the convolutional layers.
        handling_multi_graph: str = "warn"
            How to behave when dealing with multigraphs.
            Possible behaviours are:
            - "warn"
            - "raise"
            - "drop"
        node_feature_names: Optional[Union[str, List[str]]] = None
            Names of the node features.
            This is used as the layer names.
        node_type_feature_names: Optional[Union[str, List[str]]] = None
            Names of the node type features.
            This is used as the layer names.
        edge_type_feature_names: Optional[Union[str, List[str]]] = None
            Names of the edge type features.
            This is used as the layer names.
        verbose: bool = False
            Whether to show loading bars.
        """
        AbstractEdgePredictionModel.__init__(self, random_state=random_state)
        AbstractEdgeGCN.__init__(
            self,
            kernels=kernels,
            epochs=epochs,
            number_of_graph_convolution_layers=number_of_graph_convolution_layers,
            number_of_units_per_graph_convolution_layers=number_of_units_per_graph_convolution_layers,
            number_of_ffnn_body_layers=number_of_ffnn_body_layers,
            number_of_ffnn_head_layers=number_of_ffnn_head_layers,
            number_of_units_per_ffnn_body_layer=number_of_units_per_ffnn_body_layer,
            number_of_units_per_ffnn_head_layer=number_of_units_per_ffnn_head_layer,
            dropout_rate=dropout_rate,
            batch_size=batch_size,
            combiner=combiner,
            apply_norm=apply_norm,
            edge_embedding_methods=edge_embedding_methods,
            optimizer=optimizer,
            early_stopping_min_delta=early_stopping_min_delta,
            early_stopping_patience=early_stopping_patience,
            reduce_lr_min_delta=reduce_lr_min_delta,
            reduce_lr_patience=reduce_lr_patience,
            early_stopping_monitor=early_stopping_monitor,
            early_stopping_mode=early_stopping_mode,
            reduce_lr_monitor=reduce_lr_monitor,
            reduce_lr_mode=reduce_lr_mode,
            reduce_lr_factor=reduce_lr_factor,
            use_class_weights=False,
            use_edge_metrics=use_edge_metrics,
            random_state=random_state,
            use_node_embedding=use_node_embedding,
            node_embedding_size=node_embedding_size,
            use_node_type_embedding=use_node_type_embedding,
            node_type_embedding_size=node_type_embedding_size,
            use_edge_type_embedding=use_edge_type_embedding,
            edge_type_embedding_size=edge_type_embedding_size,
            residual_convolutional_layers=residual_convolutional_layers,
            siamese_node_feature_module=siamese_node_feature_module,
            handling_multi_graph=handling_multi_graph,
            node_feature_names=node_feature_names,
            node_type_feature_names=node_type_feature_names,
            edge_type_feature_names=edge_type_feature_names,
            verbose=verbose,
        )
        self._number_of_batches_per_epoch = number_of_batches_per_epoch
        self._avoid_false_negatives = avoid_false_negatives
        self._training_unbalance_rate = training_unbalance_rate
        self._training_sample_only_edges_with_heterogeneous_node_types = (
            training_sample_only_edges_with_heterogeneous_node_types
        )

    def parameters(self) -> Dict[str, Any]:
        """Returns parameters used for this model."""
        removed = [
            "use_class_weights",
        ]
        return dict(
            **{
                key: value
                for key, value in AbstractEdgeGCN.parameters(self).items()
                if key not in removed
            },
            training_unbalance_rate=self._training_unbalance_rate,
            training_sample_only_edges_with_heterogeneous_node_types=self._training_sample_only_edges_with_heterogeneous_node_types,
        )

    def _get_model_prediction_input(
        self,
        graph: Graph,
        support: Graph,
        node_features: Optional[List[np.ndarray]],
        node_type_features: Optional[List[np.ndarray]],
        edge_type_features: Optional[List[np.ndarray]],
        edge_features: Optional[
            Union[Type[AbstractEdgeFeature], List[Type[AbstractEdgeFeature]]]
        ],
    ) -> GCNEdgePredictionSequence:
        """Returns dictionary with class weights."""
        return GCNEdgePredictionSequence(
            graph,
            support=support,
            kernels=self.convert_graph_to_kernels(support),
            batch_size=self.get_batch_size_from_graph(graph),
            node_features=node_features,
            return_node_ids=self._use_node_embedding,
            return_edge_node_ids=self._use_node_embedding or self.has_kernels(),
            return_node_types=self._use_node_type_embedding,
            return_edge_types=self._use_edge_type_embedding,
            node_type_features=node_type_features,
            edge_type_features=edge_type_features,
            use_edge_metrics=self._use_edge_metrics,
            edge_features=edge_features,
        )

    def _get_model_training_input(
        self,
        graph: Graph,
        support: Graph,
        node_features: Optional[List[np.ndarray]],
        node_type_features: Optional[List[np.ndarray]],
        edge_type_features: Optional[List[np.ndarray]],
        edge_features: Optional[
            Union[Type[AbstractEdgeFeature], List[Type[AbstractEdgeFeature]]]
        ],
    ) -> GCNEdgePredictionTrainingSequence:
        """Returns training input tuple."""
        return GCNEdgePredictionTrainingSequence(
            graph,
            kernels=self.convert_graph_to_kernels(support),
            support=support,
            batch_size=self.get_batch_size_from_graph(graph),
            number_of_batches_per_epoch=self._number_of_batches_per_epoch,
            node_features=node_features,
            edge_features=edge_features,
            return_node_ids=self._use_node_embedding,
            return_edge_node_ids=self._use_node_embedding or self.has_kernels(),
            return_node_types=self._use_node_type_embedding,
            return_edge_types=self._use_edge_type_embedding,
            node_type_features=node_type_features,
            edge_type_features=edge_type_features,
            use_edge_metrics=self._use_edge_metrics,
            avoid_false_negatives=self._avoid_false_negatives,
            negative_samples_rate=self._training_unbalance_rate
            / (self._training_unbalance_rate + 1.0),
            sample_only_edges_with_heterogeneous_node_types=self._training_sample_only_edges_with_heterogeneous_node_types,
            random_state=self._random_state,
        )

    def get_output_classes(self, graph: Graph) -> int:
        """Returns number of output classes."""
        return 1

    @classmethod
    def can_use_edge_types(cls) -> bool:
        """Returns whether the model can optionally use edge types."""
        return True

    def _predict_proba(
        self,
        graph: Graph,
        support: Optional[Graph] = None,
        node_features: Optional[List[np.ndarray]] = None,
        node_type_features: Optional[List[np.ndarray]] = None,
        edge_type_features: Optional[List[np.ndarray]] = None,
        edge_features: Optional[
            Union[
                Type[AbstractEdgeFeature],
                List[Union[Type[AbstractEdgeFeature], np.ndarray]],
            ]
        ] = None,
    ) -> np.ndarray:
        """Run predictions on the provided graph."""
        predictions = super()._predict_proba(
            graph,
            support=support,
            node_features=node_features,
            node_type_features=node_type_features,
            edge_type_features=edge_type_features,
            edge_features=edge_features,
        )
        # The model will padd the predictions with a few zeros
        # in order to run the GCN portion of the model, which
        # always requires a batch size equal to the nodes number.
        return predictions[: graph.get_number_of_directed_edges()]

    @classmethod
    def smoke_test_parameters(cls) -> Dict[str, Any]:
        """Returns parameters for smoke test."""
        return dict(
            number_of_batches_per_epoch=1,
            **super().smoke_test_parameters(),
        )

    def is_using_edge_types(self) -> bool:
        """Returns whether the model is parametrized to use edge types."""
        return self._use_edge_type_embedding or super().is_using_edge_types()

    @classmethod
    def requires_edge_types(cls) -> bool:
        """Returns whether the model requires edge types."""
        return False

    @classmethod
    def requires_edge_type_features(cls) -> bool:
        return False

    @classmethod
    def requires_edge_features(cls) -> bool:
        return False

    @classmethod
    def can_use_edge_type_features(cls) -> bool:
        return True

    @classmethod
    def can_use_edge_features(cls) -> bool:
        return True

    @classmethod
    def model_name(cls) -> str:
        return "Everything Bagel GCN"


"""Keras Sequence for Open-world assumption GCN."""
from typing import List, Optional, Tuple

import numpy as np
import tensorflow as tf
from ensmallen import Graph  # pylint: disable=no-name-in-module
from keras_mixed_sequence import Sequence

from embiggen.sequences.tensorflow_sequences.gcn_edge_prediction_sequence import (
    GCNEdgePredictionSequence,
)
from embiggen.utils import AbstractEdgeFeature


class GCNEdgePredictionTrainingSequence(Sequence):
    """Keras Sequence for running Neural Network on graph edge prediction."""

    def __init__(
        self,
        graph: Graph,
        kernels: Optional[List[tf.SparseTensor]],
        batch_size: int,
        number_of_batches_per_epoch: Optional[int],
        support: Optional[Graph] = None,
        node_features: Optional[List[np.ndarray]] = None,
        node_type_features: Optional[List[np.ndarray]] = None,
        edge_type_features: Optional[List[np.ndarray]] = None,
        edge_features: Optional[List[AbstractEdgeFeature]] = None,
        return_node_ids: bool = False,
        return_edge_node_ids: bool = True,
        return_node_types: bool = False,
        return_edge_types: bool = False,
        use_edge_metrics: bool = False,
        negative_samples_rate: float = 0.5,
        avoid_false_negatives: bool = False,
        graph_to_avoid: Graph = None,
        sample_only_edges_with_heterogeneous_node_types: bool = False,
        random_state: int = 42,
    ):
        """Create new Open-world assumption GCN training sequence for edge prediction.

        Parameters
        --------------------------------
        graph: Graph,
            The graph from which to sample the edges.
        kernels: Optional[List[tf.SparseTensor]]
            The kernel to be used for the convolutions.
        batch_size: int
            The batch size to use.
        number_of_batches_per_epoch: Optional[int]
            The number of batches to use per epoch.
            If None, the number of batches will be equal to the number of edges.
        support: Optional[Graph] = None
            The graph to use to compute the edge metrics.
        node_features: Optonal[List[np.ndarray]] = None
            The node features to be used.
        node_type_features: Optional[List[np.ndarray]]
            The node type features to be used.
            For instance, these could be BERT embeddings of the
            description of the node types.
            When the graph has multilabel node types,
            we will average the features.
        edge_type_features: Optional[List[np.ndarray]] = None
            The edge type features to be used.
            For instance, these could be BERT embeddings of the
            description of the edge types.
        edge_features: Optional[List[AbstractEdgeFeature]] = None,
            The edge features to be used.
        return_node_ids: bool = False
            Whether to return the node IDs.
            These are needed when a node embedding layer is used.
        return_edge_node_ids: bool = True
            Whether to return the edge node IDs.
        return_node_types: bool = False,
            Whether to return the node type IDs.
            These are needed when a node type embedding layer is used.
        return_edge_types: bool = False,
            Whether to return the edge type IDs.
            These are needed when an edge type embedding layer is used.
        use_edge_metrics: bool = False
            Whether to return the edge metrics.
        negative_samples_rate: float = 0.5
            Factor of negatives to use in every batch.
            For example, with a batch size of 128 and negative_samples_rate equal
            to 0.5, there will be 64 positives and 64 negatives.
        avoid_false_negatives: bool = False,
            Whether to filter out false negatives.
            By default False.
            Enabling this will slow down the batch generation while (likely) not
            introducing any significant gain to the model performance.
        graph_to_avoid: Graph = None,
            Graph to avoid when generating the edges.
            This can be the validation component of the graph, for example.
            More information to how to generate the holdouts is available
            in the Graph package.
        sample_only_edges_with_heterogeneous_node_types: bool = False
            Whether to only sample edges between heterogeneous node types.
            This may be useful when training a model to predict between
            two portions in a bipartite graph.
        random_state: int = 42,
            The random_state to use to make extraction reproducible.
        """

        self._negative_samples_rate = negative_samples_rate
        self._avoid_false_negatives = avoid_false_negatives
        self._graph_to_avoid = graph_to_avoid
        self._random_state = random_state
        self._sample_only_edges_with_heterogeneous_node_types = (
            sample_only_edges_with_heterogeneous_node_types
        )
        self._current_index = 0

        self._prediction_sequence = GCNEdgePredictionSequence(
            graph=graph,
            support=support,
            kernels=kernels,
            batch_size=batch_size,
            node_features=node_features,
            node_type_features=node_type_features,
            edge_type_features=edge_type_features,
            edge_features=edge_features,
            return_node_ids=return_node_ids,
            return_edge_node_ids=return_edge_node_ids,
            return_node_types=return_node_types,
            return_edge_types=return_edge_types,
            use_edge_metrics=use_edge_metrics,
        )

        if number_of_batches_per_epoch is None:
            sample_number = graph.get_number_of_edges()
        else:
            sample_number = number_of_batches_per_epoch * batch_size

        super().__init__(
            sample_number=sample_number,
            batch_size=batch_size,
        )

    def get_node_features(
        self,
        sources: Optional[np.ndarray] = None,
        destinations: Optional[np.ndarray] = None,
    ) -> Tuple[np.ndarray]:
        """Return node features."""
        return self._prediction_sequence.get_node_features(
            sources=sources, destinations=destinations
        )

    def use_edge_metrics(self) -> bool:
        """Return whether to use edge metrics."""
        return self._prediction_sequence.use_edge_metrics()

    def get_graph(self) -> Graph:
        """Return graph."""
        return self._prediction_sequence.get_graph()

    def get_support(self) -> Graph:
        """Return support."""
        return self._prediction_sequence.get_support()

    def get_kernel(self) -> tf.SparseTensor:
        """Return kernel."""
        return self._prediction_sequence.get_kernel()

    def has_kernels(self) -> bool:
        """Return whether kernels are available."""
        return self._prediction_sequence.has_kernels()

    def return_edge_types(self) -> bool:
        """Return whether to return edge types."""
        return self._prediction_sequence.return_edge_types()

    def has_edge_type_features(self) -> bool:
        """Return whether to return edge type features."""
        return self._prediction_sequence.has_edge_type_features()
    
    def return_edge_node_ids(self) -> bool:
        """Return whether to return edge node IDs."""
        return self._prediction_sequence.return_edge_node_ids()

    def __getitem__(self, idx: int) -> Tuple[np.ndarray, np.ndarray]:
        """Return batch corresponding to given index.

        Parameters
        ---------------
        idx: int,
            Index corresponding to batch to be returned.

        Returns
        ---------------
        Return Tuple containing X and Y numpy arrays corresponding to given batch index.
        """
        (
            sources,
            _,
            destinations,
            _,
            edge_types,
            edge_metrics,
            labels,
        ) = self.get_graph().get_edge_prediction_mini_batch(
            (self._random_state + idx) * (1 + self.elapsed_epochs),
            return_node_types=False,
            return_edge_types=self.return_edge_types() or self.has_edge_type_features(),
            return_edge_metrics=self.use_edge_metrics(),
            batch_size=self.batch_size,
            negative_samples_rate=self._negative_samples_rate,
            avoid_false_negatives=self._avoid_false_negatives,
            sample_only_edges_with_heterogeneous_node_types=self._sample_only_edges_with_heterogeneous_node_types,
            support=self.get_support(),
            graph_to_avoid=self._graph_to_avoid,
        )

        edge_features: Tuple[
            np.ndarray
        ] = self._prediction_sequence.get_edge_features_from_edge_node_ids(
            sources, destinations
        )

        edge_type_features = (
            self._prediction_sequence.get_edge_type_features(edge_types)
            if self.has_edge_type_features()
            else []
        )

        # We need to reshape the sources and destinations to be
        # column vectors, as expected by the model input shapes.
        sources = sources.reshape((-1, 1))
        destinations = destinations.reshape((-1, 1))

        if self.return_edge_types():
            edge_types = edge_types.reshape((-1, 1))

        return (
            tuple(
                [
                    value
                    for value in (
                        sources if self.return_edge_node_ids() else None,
                        destinations if self.return_edge_node_ids() else None,
                        (edge_types if self.return_edge_types() else None),
                        edge_metrics,
                        *edge_features,
                        *edge_type_features,
                        *(
                            self.get_node_features()
                            if self.has_kernels()
                            else self.get_node_features(
                                sources=sources, destinations=destinations
                            )
                        ),
                    )
                    if value is not None
                ]
            ),
            labels,
        )



import pytest
from embiggen.utils import EmbeddingResult
from unittest import TestCase
import numpy as np


class TestEmbeddingResult(TestCase):

    def setUp(self):
        pass

    def test_construction(self):
        nr = EmbeddingResult(
            embedding_method_name="Test",
            node_embeddings=np.random.uniform(size=(100, 10))
        )
        nr.get_all_node_embedding()
        nr.get_node_embedding_from_index(0)

        self.assertEqual("Test", nr.embedding_method_name)

        nr.dump()

        with pytest.raises(ValueError):
            nr.get_node_embedding_from_index(1)

        nr = EmbeddingResult(
            embedding_method_name="Test",
            edge_embeddings=np.random.uniform(size=(100, 10))
        )
        nr.get_all_edge_embedding()
        nr.get_edge_embedding_from_index(0)

        with pytest.raises(ValueError):
            nr.get_edge_embedding_from_index(1)

        nr = EmbeddingResult(
            embedding_method_name="Test",
            node_type_embeddings=np.random.uniform(size=(100, 10))
        )
        nr.get_all_node_type_embeddings()
        nr.get_node_type_embedding_from_index(0)

        with pytest.raises(ValueError):
            nr.get_node_type_embedding_from_index(1)

        nr = EmbeddingResult(
            embedding_method_name="Test",
            edge_type_embeddings=np.random.uniform(size=(100, 10))
        )
        nr.get_all_edge_type_embeddings()
        nr.get_edge_type_embedding_from_index(0)

        with pytest.raises(ValueError):
            nr.get_edge_type_embedding_from_index(1)

        with pytest.raises(ValueError):
            EmbeddingResult(
                embedding_method_name="Test",
                edge_type_embeddings="hu"
            )

        with pytest.raises(ValueError):
            EmbeddingResult(
                embedding_method_name="Test",
                edge_type_embeddings=np.random.uniform(size=(0, 10))
            )

        with pytest.raises(ValueError):
            EmbeddingResult(
                embedding_method_name="Test",
                edge_type_embeddings=np.full(shape=(10, 10), fill_value=np.nan)
            )

    def test_check_exceptions(self):
        node_embedding = EmbeddingResult(
            embedding_method_name="Test",
        )

        with pytest.raises(ValueError):
            node_embedding.get_all_node_embedding()

        with pytest.raises(ValueError):
            node_embedding.get_all_edge_embedding()

        with pytest.raises(ValueError):
            node_embedding.get_all_node_type_embeddings()
        
        with pytest.raises(ValueError):
            node_embedding.get_all_edge_type_embeddings()


"""Unit test class for testing whether edge features are correctly handled in the pipeline with multiple graphs"""
from unittest import TestCase
from embiggen.edge_prediction import edge_prediction_evaluation, DecisionTreeEdgePrediction, RandomForestEdgePrediction
from ensmallen.datasets.linqs import Cora
from ensmallen.datasets.kgobo import CIO
from embiggen.embedders.ensmallen_embedders.hyper_sketching import HyperSketching


class TestEdgeFeaturesInMultipleGraphPipeline(TestCase):
    """Unit test class for testing whether edge features are correctly handled in the pipeline with multiple graphs."""

    def setUp(self):
        """Setup objects for running tests on GraphTransformer objects class."""
        self._first_graph = Cora()
        self._second_graph = CIO()


    def test_edge_features_from_string_in_multiple_graph_pipeline(self):
        """Tests whether running from edge features defined as string works in the pipeline with multiple graphs."""
        edge_prediction_evaluation(
            holdouts_kwargs=dict(train_size=0.8),
            models=[
                "Decision Tree Classifier",
                "Random Forest Classifier"
                ],
            edge_features="HyperSketching",
            evaluation_schema="Monte Carlo",
            graphs=[
                self._first_graph,
                self._second_graph
            ],
            number_of_holdouts=2,
            smoke_test=True,
            verbose=False,
        )

    def test_edge_features_from_instance_in_multiple_graph_pipeline(self):
        """Tests whether running from edge features defined as instance works in the pipeline with multiple graphs."""
        edge_prediction_evaluation(
            holdouts_kwargs=dict(train_size=0.8),
            models=[
                "Decision Tree Classifier",
                "Random Forest Classifier"
            ],
            edge_features=HyperSketching(),
            evaluation_schema="Monte Carlo",
            graphs=[
                self._first_graph,
                self._second_graph
            ],
            number_of_holdouts=2,
            smoke_test=True,
            verbose=False,
        )

    def test_edge_features_from_instance_in_multiple_graph_pipeline_with_model_instance(self):
        """Tests whether running from edge features defined as instance works in the pipeline with multiple graphs."""
        edge_prediction_evaluation(
            holdouts_kwargs=dict(train_size=0.8),
            models=[
                RandomForestEdgePrediction(),
                DecisionTreeEdgePrediction()
            ],
            edge_features=HyperSketching(),
            evaluation_schema="Monte Carlo",
            graphs=[
                self._first_graph,
                self._second_graph
            ],
            number_of_holdouts=2,
            smoke_test=True,
            verbose=False,
        )



"""Keras Sequence for running Neural Network on graph edge prediction."""
import numpy as np
from ensmallen import Graph  # pylint: disable=no-name-in-module


class EdgePredictionSequence:
    """Keras Sequence for running Neural Network on graph edge prediction."""

    def __init__(
        self,
        graph: Graph,
        support: Graph,
        return_node_types: bool,
        return_edge_types: bool,
        use_edge_metrics: bool,
        batch_size: int = 2**10,
    ):
        """Create new EdgePredictionSequence object.

        Parameters
        --------------------------------
        graph: Graph
            The graph whose edges are to be predicted.
        support: Graph
            The graph that was used while training the current
            edge prediction model.
        return_node_types: bool
            Whether to return the node types.
        return_edge_types: bool
            Whether to return the edge types.
        use_edge_metrics: bool = False
            Whether to return the edge metrics.
        batch_size: int = 2**10,
            The batch size to use.
        """
        if not graph.has_compatible_node_vocabularies(support):
            raise ValueError(
                f"The provided graph {graph.get_name()} does not have a node vocabulary "
                "that is compatible with the provided graph used in training."
            )
        
        if support is None:
            support = graph

        self._graph = graph
        self._support = support
        self._return_node_types = return_node_types
        self._return_edge_types = return_edge_types
        self._use_edge_metrics = use_edge_metrics
        self._batch_size = batch_size

    def __len__(self) -> int:
        """Returns length of sequence."""
        return int(np.ceil(self._graph.get_number_of_directed_edges() / self._batch_size))
    
    def use_edge_metrics(self) -> bool:
        """Return whether to use edge metrics."""
        return self._use_edge_metrics
    
    def return_node_types(self) -> bool:
        """Return whether to return node types."""
        return self._return_node_types
    
    def return_edge_types(self) -> bool:
        """Return whether to return edge types."""
        return self._return_edge_types
    
    def get_graph(self) -> Graph:
        """Return graph."""
        return self._graph
    
    def get_support(self) -> Graph:
        """Return support graph."""
        return self._support
    

    def __getitem__(self, idx: int):
        """Return batch corresponding to given index.

        Parameters
        ---------------
        idx: int,
            Index corresponding to batch to be returned.

        Returns
        ---------------
        """
        return (tuple([
            value
            for value in self._support.get_edge_prediction_chunk_mini_batch(
                idx,
                graph=self._graph,
                batch_size=self._batch_size,
                return_node_types=self._return_node_types,
                return_edge_types=self._return_edge_types,
                return_edge_metrics=self.use_edge_metrics(),
            )
            if value is not None
        ]),)



"""Keras Sequence for edge-label prediction GNN and GCN."""
from typing import List, Optional, Union, Type, Tuple

import numpy as np
import tensorflow as tf
from ensmallen import Graph  # pylint: disable=no-name-in-module
from keras_mixed_sequence import VectorSequence, Sequence
from embiggen.utils import AbstractEdgeFeature
from embiggen.sequences.tensorflow_sequences.gcn_edge_prediction_training_sequence import GCNEdgePredictionTrainingSequence


class GCNEdgeLabelPredictionSequence(Sequence):
    """Keras Sequence for running Neural Network on graph edge-label prediction."""

    def __init__(
        self,
        graph: Graph,
        support: Graph,
        kernels: Optional[List[tf.SparseTensor]],
        batch_size: int,
        return_node_types: bool = False,
        return_node_ids: bool = False,
        return_edge_node_ids: bool = True,
        node_features: Optional[List[np.ndarray]] = None,
        node_type_features: Optional[List[np.ndarray]] = None,
        edge_features: Optional[Union[np.ndarray, Type[AbstractEdgeFeature], List[Union[Type[AbstractEdgeFeature], np.ndarray]]]] = None,
        use_edge_metrics: bool = False,
    ):
        """Create new Open-world assumption GCN training sequence for edge prediction.

        Parameters
        --------------------------------
        graph: Graph,
            The graph from which to sample the edges.
        support: Graph
            The graph to be used for the topological metrics.
        kernels: Optional[List[tf.SparseTensor]]
            The kernel to be used for the convolutions.
        batch_size: int
            The batch size to use.
        return_node_types: bool = False
            Whether to return the node types.
        return_edge_types: bool = False
            Whether to return the edge types.
        return_node_ids: bool = False
            Whether to return the node IDs.
            These are needed when an embedding layer is used.
        return_edge_node_ids: bool = True
            Whether to return the edge node IDs.
            These are needed when an embedding layer is used.
        node_features: List[np.ndarray]
            The node features to be used.
        node_type_features: Optional[List[np.ndarray]]
            The node type features to be used.
            For instance, these could be BERT embeddings of the
            description of the node types.
            When the graph has multilabel node types,
            we will average the features.
        edge_features: Optional[List[np.ndarray]] = None,
            The edge features to be used.
            For instance, these could be BERT embeddings of the
            description of the edges.
        use_edge_metrics: bool = False
            Whether to return the edge metrics.
        """
        super().__init__(
            sample_number=graph.get_number_of_edges(),
            batch_size=batch_size,
        )

        # We use the GCN edge prediction training sequence to avoid
        # duplicating the complex logic that handles the computation
        # of the kernels, node and node type metrics.
        self._gcn_edge_prediction_training_sequence = GCNEdgePredictionTrainingSequence(
            graph=graph,
            support=support,
            kernels=kernels,
            batch_size=batch_size,
            number_of_batches_per_epoch=None,
            return_node_types=return_node_types,
            return_node_ids=return_node_ids,
            return_edge_node_ids=return_edge_node_ids,
            node_features=node_features,
            node_type_features=node_type_features,
        )

        # Differently from the GCN edge prediction training sequence, the edge-label
        # prediction training sequence can handle the dense edge features, if provided.
        # We proceed to validate the provided edge features, if any.
        # The edge features that are supported as:
        # - One or more numpy array of shape (number_of_edges, number_of_features)
        # - One or more edge feature classes

        if edge_features is None:
            edge_features = []

        if not isinstance(edge_features, list):
            edge_features = [edge_features]

        rasterized_edge_features = []

        if use_edge_metrics:
            rasterized_edge_features.append(
                support.get_all_edge_metrics(
                    normalize=True,
                    subgraph=graph,
                )
            )

        for edge_feature in edge_features:
            if not isinstance(edge_feature, np.ndarray) and not issubclass(type(edge_feature), AbstractEdgeFeature):
                raise ValueError(
                    "The edge features should be provided as a numpy array or "
                    f"an edge feature class. Got {type(edge_feature)} instead."
                )
            if isinstance(edge_feature, np.ndarray) and len(edge_feature.shape) != 2:
                raise ValueError(
                    "The edge features should be provided as a numpy array with "
                    "shape (number_of_edges, number_of_features). "
                    f"Got {edge_feature.shape} instead."
                )
            if isinstance(edge_feature, np.ndarray) and edge_feature.shape[0] != graph.get_number_of_edges():
                raise ValueError(
                    "The edge features should be provided as a numpy array with "
                    f"shape (number_of_edges, number_of_features). "
                    f"Got {edge_feature.shape} instead, while the number of edges is {graph.get_number_of_edges()}."
                )
            if isinstance(edge_feature, np.ndarray):
                rasterized_edge_features.append(edge_feature)
            elif issubclass(type(edge_feature), AbstractEdgeFeature):
                if not edge_feature.is_fit():
                    edge_feature.fit(support=support)
                for feature in edge_feature.get_edge_feature_from_graph(
                    graph=graph,
                    support=support,
                ).values():
                    rasterized_edge_features.append(feature.reshape(feature.shape[0], -1))

        self._edge_features = [
            VectorSequence(
                rasterized_edge_feature,
                batch_size=self.batch_size,
            )
            for rasterized_edge_feature in rasterized_edge_features
        ]
        self._sources = VectorSequence(
            graph.get_source_node_ids(graph.is_directed()),
            batch_size=self.batch_size,
        )
        self._destinations = VectorSequence(
            graph.get_destination_node_ids(graph.is_directed()),
            batch_size=self.batch_size,
        )
        
    def get_node_features(
        self,
        sources: Optional[np.ndarray]=None,
        destinations: Optional[np.ndarray]=None
    ) -> Tuple[np.ndarray]:
        """Return node features."""
        return self._gcn_edge_prediction_training_sequence.get_node_features(
            sources=sources,
            destinations=destinations
        )
    
    def has_kernels(self) -> bool:
        """Return whether the sequence has kernels."""
        return self._gcn_edge_prediction_training_sequence.has_kernels()
    
    def return_edge_node_ids(self) -> bool:
        """Return whether the sequence returns edge node IDs."""
        return self._gcn_edge_prediction_training_sequence.return_edge_node_ids()

    def __getitem__(self, idx: int):
        """Return batch corresponding to given index.

        Parameters
        ---------------
        idx: int,
            Index corresponding to batch to be returned.

        Returns
        ---------------
        Return Tuple containing X and Y numpy arrays corresponding to given batch index.
        """
        edge_features = [
            edge_feature[idx]
            for edge_feature in self._edge_features
        ]
        sources = self._sources[idx]
        destinations = self._destinations[idx]

        # If this last batch is smaller than the batch size, we need to pad it.
        # This is necessary because in GCNs, the batch size is fixed to the number of nodes.
        delta = self.batch_size - sources.shape[0]
        if delta > 0 and self.has_kernels():
            edge_features = [
                np.pad(edge_feature, [(0, delta), (0, 0)])
                for edge_feature in edge_features
            ]
            sources = np.pad(sources, (0, delta))
            destinations = np.pad(destinations, (0, delta))

        assert sources.shape[0] == destinations.shape[0]

        for edge_feature in edge_features:
            assert edge_feature.shape[0] == sources.shape[0]

        # We need to reshape the source and destination nodes to be of shape
        # (batch_size, 1) instead of (batch_size, ). This is necessary because
        # this way these features match exactly the shape of the expected input
        # of the model.

        sources = sources.reshape(-1, 1)
        destinations = destinations.reshape(-1, 1)

        return (
            (
                *((sources,) if self.return_edge_node_ids() else ()),
                *((destinations,) if self.return_edge_node_ids() else ()),
                *edge_features,
                *(self.get_node_features() if self.has_kernels() else self.get_node_features(
                    sources=sources,
                    destinations=destinations
                ))
            ),
        )



"""Submodule wrapping Logistic Regression Cross Validator for Node-label prediction."""
from typing import Dict, Any, Optional
from sklearn.linear_model import LogisticRegressionCV
from embiggen.node_label_prediction.node_label_prediction_sklearn.sklearn_node_label_prediction_adapter import (
    SklearnNodeLabelPredictionAdapter,
)
from embiggen.utils import normalize_kwargs


class LogisticRegressionCVNodeLabelPrediction(SklearnNodeLabelPredictionAdapter):
    """Create wrapper over Sklearn Logistic Regression Cross Validator for Node-label prediction."""

    def __init__(
        self,
        Cs: int = 10,
        fit_intercept: bool = True,
        cv: int = 10,
        dual: bool = False,
        penalty: str = "l2",
        scoring: str = "f1_macro",
        solver: str = "lbfgs",
        tol: float = 1e-4,
        max_iter: int = 100,
        class_weight: Optional[str] = "balanced",
        n_jobs: int = -1,
        verbose: int = 0,
        refit: bool = True,
        intercept_scaling: float = 1.0,
        multi_class: str = "auto",
        random_state: int = 42,
        l1_ratios=None,
    ):
        """Create the Logistic Regression Cross Validator for Node-label prediction."""
        self._kwargs = normalize_kwargs(
            self,
            dict(
                Cs=Cs,
                fit_intercept=fit_intercept,
                cv=cv,
                dual=dual,
                penalty=penalty,
                scoring=scoring,
                solver=solver,
                tol=tol,
                max_iter=max_iter,
                class_weight=class_weight,
                n_jobs=n_jobs,
                verbose=verbose,
                refit=refit,
                intercept_scaling=intercept_scaling,
                multi_class=multi_class,
                l1_ratios=l1_ratios,
            )
        )

        super().__init__(
            LogisticRegressionCV(**self._kwargs, random_state=random_state),
            random_state=random_state,
        )

    def parameters(self) -> Dict[str, Any]:
        """Returns parameters used for this model."""
        return {**super().parameters(), **self._kwargs}

    @classmethod
    def smoke_test_parameters(cls) -> Dict[str, Any]:
        """Returns parameters for smoke test."""
        return dict(cv=2)

    @classmethod
    def model_name(cls) -> str:
        return "Logistic Regression Cross Validator"
    
    @classmethod
    def supports_multilabel_prediction(cls) -> bool:
        """Returns whether the model supports multilabel prediction."""
        return False



"""Keras Sequence for edge-label prediction GNN and GCN."""
from typing import List, Optional, Union, Type

import numpy as np
import tensorflow as tf
from ensmallen import Graph  # pylint: disable=no-name-in-module
from keras_mixed_sequence import VectorSequence
from embiggen.utils import AbstractEdgeFeature
from embiggen.sequences.tensorflow_sequences.gcn_edge_label_prediction_sequence import GCNEdgeLabelPredictionSequence


class GCNEdgeLabelPredictionTrainingSequence(GCNEdgeLabelPredictionSequence):
    """Keras Sequence for running Neural Network on graph edge-label prediction."""

    def __init__(
        self,
        graph: Graph,
        support: Graph,
        kernels: Optional[List[tf.SparseTensor]],
        batch_size: int,
        return_node_types: bool = False,
        return_node_ids: bool = False,
        return_edge_node_ids: bool = True,
        node_features: Optional[List[np.ndarray]] = None,
        node_type_features: Optional[List[np.ndarray]] = None,
        edge_features: Optional[Union[np.ndarray, Type[AbstractEdgeFeature], List[Union[Type[AbstractEdgeFeature], np.ndarray]]]] = None,
        use_edge_metrics: bool = False,
    ):
        """Create new Open-world assumption GCN training sequence for edge prediction.

        Parameters
        --------------------------------
        graph: Graph,
            The graph from which to sample the edges.
        support: Graph
            The graph to be used for the topological metrics.
        kernels: Optional[List[tf.SparseTensor]]
            The kernel to be used for the convolutions.
        batch_size: int
            The batch size to use.
        return_node_types: bool = False
            Whether to return the node types.
        return_edge_types: bool = False
            Whether to return the edge types.
        return_node_ids: bool = False
            Whether to return the node IDs.
            These are needed when an embedding layer is used.
        return_edge_node_ids: bool = True
            Whether to return the edge node IDs.
            These are needed when an embedding layer is used.
        node_features: List[np.ndarray]
            The node features to be used.
        node_type_features: Optional[List[np.ndarray]]
            The node type features to be used.
            For instance, these could be BERT embeddings of the
            description of the node types.
            When the graph has multilabel node types,
            we will average the features.
        edge_features: Optional[List[np.ndarray]] = None,
            The edge features to be used.
            For instance, these could be BERT embeddings of the
            description of the edges.
            When the graph has multilabel edges,
            we will average the features.
        use_edge_metrics: bool = False
            Whether to return the edge metrics.
        """
        super().__init__(
            graph=graph,
            support=support,
            kernels=kernels,
            batch_size=batch_size,
            return_node_types=return_node_types,
            return_node_ids=return_node_ids,
            return_edge_node_ids=return_edge_node_ids,
            node_features=node_features,
            node_type_features=node_type_features,
            edge_features=edge_features,
            use_edge_metrics=use_edge_metrics
        )

        if graph.is_directed():
            edge_types = graph.get_imputed_directed_edge_type_ids(
                imputation_edge_type_id=0
            )
            mask = graph.get_directed_edges_with_known_edge_types_mask()
        else:
            edge_types = graph.get_imputed_upper_triangular_edge_type_ids(
                imputation_edge_type_id=0
            )
            mask = graph.get_upper_triangular_known_edge_types_mask()

        if not isinstance(edge_types, np.ndarray):
            raise RuntimeError(
                "The edge types should be a numpy array, "
                f"found {type(edge_types)} instead. "
                "This is likely an Ensmallen bug, "
                "please open an issue at "
                "the GRAPE GitHub repository."
            )

        self._edge_types = VectorSequence(
            edge_types,
            batch_size=self._batch_size,
        )

        self._mask = VectorSequence(
            mask.astype(np.uint8),
            batch_size=self._batch_size,
        )

    def __getitem__(self, idx: int):
        """Return batch corresponding to given index.

        Parameters
        ---------------
        idx: int,
            Index corresponding to batch to be returned.

        Returns
        ---------------
        Return Tuple containing X and Y numpy arrays corresponding to given batch index.
        """
        features = super().__getitem__(idx)[0]

        edge_types = self._edge_types[idx]
        mask = self._mask[idx]

        # If this last batch is smaller than the batch size, we need to pad it.
        # This is necessary because in GCNs, the batch size is fixed to the number of nodes.
        delta = self.batch_size - edge_types.shape[0]
        if delta > 0 and self.has_kernels():
            edge_types = np.pad(edge_types, (0, delta))
            mask = np.pad(mask, (0, delta))

        return (
            features,
            edge_types,
            mask
        )



"""Module to test the Ridge Classifier models for node-label, edge-label and edge prediction."""
from embiggen.node_label_prediction import (
    CatBoostNodeLabelPrediction,
)
from embiggen.edge_prediction import (
    CatBoostEdgePrediction,
)
from embiggen.edge_label_prediction import (
    CatBoostEdgeLabelPrediction,
)
from embiggen.embedders.ensmallen_embedders.hyper_sketching import HyperSketching
from ensmallen.datasets.kgobo import MIAPA
from ensmallen import Graph
from ensmallen.datasets.linqs import Cora, get_words_data
from embiggen.embedders.ensmallen_embedders.degree_spine import DegreeSPINE


def test_catboost_node_label_prediction_smoke_test():
    """Smoke test for the Ridge Classifier node-label prediction model."""
    if not CatBoostNodeLabelPrediction.is_available():
        return
    graph, data = get_words_data(Cora())
    graph = graph.remove_singleton_nodes()
    red = graph.set_all_node_types("red")
    binary_graph = (
        red | MIAPA().remove_edge_types().set_all_node_types("blue")
    ).add_selfloops()
    another_binary_graph = Graph.generate_random_connected_graph(
        random_state=100,
        number_of_nodes=100,
        node_type="red",
    ) | Graph.generate_random_connected_graph(
        minimum_node_id=100,
        random_state=1670,
        number_of_nodes=100,
        node_type="blue",
    )

    for g, data in (
        (graph, data),
        (binary_graph, None),
        (another_binary_graph, None),
    ):
        if data is None:
            data = DegreeSPINE().into_smoke_test().fit_transform(g)
        model = CatBoostNodeLabelPrediction().into_smoke_test()
        model.fit(g, node_features=data)
        model.predict(g, node_features=data)
        model.predict_proba(g, node_features=data)


def test_catboost_edge_prediction_smoke_test():
    """Smoke test for the Ridge Classifier edge-label prediction model."""
    if not CatBoostEdgePrediction.is_available():
        return
    graph, data = get_words_data(Cora())
    graph = graph.remove_singleton_nodes()
    red = graph.set_all_node_types("red")
    multilabel_graph = (
        Cora().remove_edge_weights().remove_edge_types() | red
    ).add_selfloops()
    binary_graph = (
        red | MIAPA().remove_edge_types().set_all_node_types("blue")
    ).add_selfloops()
    another_binary_graph = Graph.generate_random_connected_graph(
        random_state=100,
        number_of_nodes=100,
        node_type="red",
    ) | Graph.generate_random_connected_graph(
        minimum_node_id=100,
        random_state=1670,
        number_of_nodes=100,
        node_type="blue",
    )

    for g, data in (
        (graph, data),
        (multilabel_graph, None),
        (binary_graph, None),
        (another_binary_graph, None),
    ):
        if data is None:
            data = DegreeSPINE().into_smoke_test().fit_transform(g)
        model = CatBoostEdgePrediction(
            edge_embedding_methods=["Concatenate", "Hadamard"],
        ).into_smoke_test()
        model.fit(g, node_features=data, edge_features=HyperSketching())
        model.predict(g, node_features=data, edge_features=HyperSketching())
        model.predict_proba(g, node_features=data, edge_features=HyperSketching())


def test_catboost_edge_label_prediction_smoke_test():
    """Smoke test for the Ridge Classifier edge-label prediction model."""
    if not CatBoostEdgeLabelPrediction.is_available():
        return
    graph, data = get_words_data(Cora())
    red = graph.set_all_edge_types("red")
    blue = MIAPA().remove_singleton_nodes().set_all_edge_types("blue")
    binary = red | blue

    for g, data in ((binary, None),):
        if data is None:
            data = DegreeSPINE().into_smoke_test().fit_transform(g)
        model = CatBoostEdgeLabelPrediction(
            edge_embedding_methods=["Concatenate", "Hadamard"],
        ).into_smoke_test()
        model.fit(g, node_features=data, edge_features=HyperSketching())
        model.predict(g, node_features=data, edge_features=HyperSketching())
        model.predict_proba(g, node_features=data, edge_features=HyperSketching())



"""This module provides the Graph Convolution feature preprocessor class."""
import warnings
from typing import Any, Dict, List, Union

import numpy as np
import pandas as pd
from ensmallen import Graph, models

from embiggen.utils.abstract_models.abstract_feature_preprocessor import \
    AbstractFeaturePreprocessor
from embiggen.utils.abstract_models.embedding_result import EmbeddingResult


class GraphConvolution(AbstractFeaturePreprocessor):
    """Graph Convolution feature preprocessor class."""

    def __init__(
        self,
        number_of_convolutions: int = 2,
        concatenate_features: bool = False,
        transpose: bool = False,
        normalize_rows: bool = False,
        dtype: str = "f32",
        path: Union[str, List[str]] = None,
    ):
        """Create new Graph Convolution feature preprocessor.

        Parameters
        -------------------------
        number_of_convolutions: int = 2
            The number of convolutions to execute.
            By default, `2`.
        concatenate_features: bool = False
            Whether to Concatenate the features at each convolution.
            By default, `false`.
        transpose: bool = False
            Whether to transpose the features before convolving them.
            By default, `false`.
        normalize_rows: bool = True
            Whether to normalize the rows of the features before convolving them.
            By default, `true`.
        dtype: str = "f32"
            The data type to use for the convolved features.
            The supported values are `f16`, `f32` and `f64`.
            By default, `f32`.
        path: Union[str, List[str]] = None
            The path(s) were to MMAP the processed features to.
            By default, `None`.
        """
        self._kwargs = dict(
            number_of_convolutions=number_of_convolutions,
            concatenate_features=concatenate_features,
            normalize_rows=normalize_rows,
            dtype=dtype,
        )
        self._transpose = transpose
        self._path = path
        self._graph_convolution = models.GraphConvolution(
            **self._kwargs
        )
        super().__init__()

    @classmethod
    def model_name(cls) -> str:
        """Return the name of the feature preprocessor."""
        return "Graph Convolution"

    @classmethod
    def library_name(cls) -> str:
        """Return the name of the library of the feature preprocessor."""
        return "Ensmallen"

    @classmethod
    def requires_edge_types(cls) -> bool:
        """Return whether the model requires edge types."""
        return False

    @classmethod
    def requires_node_types(cls) -> bool:
        """Return whether the model requires node types."""
        return False

    @classmethod
    def requires_edge_weights(cls) -> bool:
        """Return whether the model requires edge weights."""
        return False
    
    def parameters(self) -> Dict[str, Any]:
        return dict(
            **self._kwargs,
            transpose=self._transpose,
            path=self._path
        )

    def _transform(
        self,
        support: Graph,
        node_features: List[Union[pd.DataFrame, np.ndarray]],
    ) -> EmbeddingResult:
        """Transform the given node features.

        Parameters
        -------------------------
        node_features: List[Union[pd.DataFrame, np.ndarray]],
            Node feature to use to fit the transformer.
        support: Graph
            Support graph to use to transform the node features.
        """
        if self._path is not None and len(self._path) != len(node_features):
            raise ValueError(
                "The number of paths should be the same as the number of node features."
                f"Got {len(self._path)} paths and {len(node_features)} node features."
            )
        
        new_node_features = []

        for node_feature in node_features:
            if not isinstance(node_feature, np.ndarray):
                raise NotImplementedError(
                    "The node features should be provided as a numpy array. "
                    f"Got {type(node_feature)} instead."
                )
            if len(node_feature.shape) != 2:
                raise ValueError(
                    "The node features should be provided as a numpy array with "
                    "shape (number_of_nodes, number_of_features). "
                    f"Got {node_feature.shape} instead."
                )
            
            if not node_feature.data.c_contiguous:
                warnings.warn(
                    "One of the provided node features is not C contiguous. "
                    f"Specifically, the one with shape {node_feature.shape} "
                    f"and dtype {node_feature.dtype}. "
                    "This forces us to copy the data, which may be slow and "
                    "consume a lot of memory. Consider using np.ascontiguousarray "
                    "to make the array C contiguous before passing it to the "
                    "feature preprocessor."
                )
                node_feature = np.ascontiguousarray(node_feature)
            new_node_features.append(node_feature)

        node_features = new_node_features

        # We transpose the graph if requested, though the operation is skipped
        # if we are computing the transposed of an undirected graph. A warning
        # is raised in this case.
        if self._transpose:
            if graph.is_directed():
                graph = graph.to_transposed()
            else:
                warnings.warn(
                    "You are trying to compute the transposed of an undirected graph. "
                    "The transposed of an undirected graph is the same graph. "
                    "This operation is skipped."
                )

        if self._path is None:
            transposed_features = [
                self._graph_convolution.transform(
                    support=support,
                    node_features=node_feature,
                )
                for node_feature in node_features
            ]
        else:
            transposed_features = [
                self._graph_convolution.transform(
                    support=support,
                    node_feature=node_feature,
                    path=path
                )
                for node_feature, path in zip(node_features, self._path)
            ]
        return EmbeddingResult(
            embedding_method_name=self.model_name(),
            node_embeddings=transposed_features,
        )

    @classmethod
    def is_stocastic(cls) -> bool:
        """Return whether the model is stocastic."""
        return False

    @classmethod
    def can_use_edge_type_features(cls) -> bool:
        return False
    
    @classmethod
    def can_use_edge_features(cls) -> bool:
        return False


"""Node-label prediction model based on XGB."""
from typing import Dict, Any, Optional, Union, List
from xgboost import XGBClassifier
import numpy as np
from embiggen.node_label_prediction.sklearn_like_node_label_prediction_adapter import (
    SklearnLikeNodeLabelPredictionAdapter,
)


class XGBNodeLabelPrediction(SklearnLikeNodeLabelPredictionAdapter):
    """Node-label prediction model based on XGB."""

    def __init__(
        self,
        max_depth: int = 6,
        max_leaves: int = 0,
        max_bin: int = 256,
        grow_policy: Optional[str] = None,
        learning_rate: float = 0.3,
        n_estimators: int = 100,
        verbosity: int = 1,
        objective = None,
        booster: Optional[str] = None,
        tree_method: str = "auto",
        n_jobs: int = -1,
        gamma: Optional[float] = None,
        min_child_weight: float = 1.0,
        max_delta_step: float = 0.0,
        subsample: float = 1.0,
        sampling_method: str = "uniform",
        colsample_bytree: float = 1.0,
        colsample_bylevel: float = 1.0,
        colsample_bynode: float = 1.0,
        reg_alpha: float = 0.0,
        reg_lambda: float = 0.0,
        base_score: Optional[float] = None,
        missing: float = np.nan,
        num_parallel_tree: int = 1,
        monotone_constraints: Optional[Union[Dict[str, int], str]] = None,
        interaction_constraints: Optional[List[List[int]]] = None,
        importance_type: Optional[str] = None,
        gpu_id: Optional[int] = None,
        validate_parameters: bool = False,
        predictor: str = "auto",
        enable_categorical: bool = False,
        max_cat_to_onehot: Optional[int] = None,
        max_cat_threshold: Optional[int] = None,
        eval_metric: Optional[Union[str, List[str]]] = None,
        early_stopping_rounds: Optional[int] = None,
        random_state: int = 42
    ):
        """Build a XGB node-label prediction model."""
        self._kwargs = dict(
            max_depth=max_depth,
            max_leaves=max_leaves,
            max_bin=max_bin,
            grow_policy=grow_policy,
            learning_rate=learning_rate,
            n_estimators=n_estimators,
            verbosity=verbosity,
            objective=objective,
            booster=booster,
            tree_method=tree_method,
            n_jobs=n_jobs,
            gamma=gamma,
            min_child_weight=min_child_weight,
            max_delta_step=max_delta_step,
            subsample=subsample,
            sampling_method=sampling_method,
            colsample_bytree=colsample_bytree,
            colsample_bylevel=colsample_bylevel,
            colsample_bynode=colsample_bynode,
            reg_alpha=reg_alpha,
            reg_lambda=reg_lambda,
            base_score=base_score,
            missing=missing,
            num_parallel_tree=num_parallel_tree,
            monotone_constraints=monotone_constraints,
            interaction_constraints=interaction_constraints,
            importance_type=importance_type,
            gpu_id=gpu_id,
            validate_parameters=validate_parameters,
            predictor=predictor,
            enable_categorical=enable_categorical,
            max_cat_to_onehot=max_cat_to_onehot,
            max_cat_threshold=max_cat_threshold,
            eval_metric=eval_metric,
            early_stopping_rounds=early_stopping_rounds,
        )

        super().__init__(
            model_instance=XGBClassifier(
                **self._kwargs,
                random_state=random_state,
            ),
            random_state=random_state,
        )

    @classmethod
    def smoke_test_parameters(cls) -> Dict[str, Any]:
        return dict(
            n_estimators=2,
            max_depth=2,
        )

    def parameters(self) -> Dict[str, Any]:
        return dict(
            **super().parameters(),
            **self._kwargs,
        )

    @classmethod
    def model_name(cls) -> str:
        """Return the name of the model."""
        return "XGBoost"

    @classmethod
    def library_name(cls) -> str:
        """Return name of the model."""
        return "XGBoost"

    @classmethod
    def supports_multilabel_prediction(cls) -> bool:
        """Returns whether the model supports multilabel prediction."""
        return False



from typing import List, Optional, Union, Dict, Tuple
import pandas as pd
import numpy as np
from ensmallen import models, Graph


class DAGResnik:

    def __init__(self, verbose: bool = True):
        """Create new Resnik similarity model."""
        self._model = models.DAGResnik(verbose)
        self._graph = None

    def fit(
        self,
        graph: Graph,
        node_counts: Dict[str, float],
        node_frequencies: Optional[np.ndarray] = None,
    ):
        """Fit the Resnik similarity model.

        Parameters
        --------------------
        graph: Graph
            The graph to run similarities on.
        node_counts: Dict[str, float]
            Counts to compute the terms frequencies.
        node_frequencies: Optional[np.ndarray] = None
            Optional vector of node frequencies.
        """
        self._graph = graph
        self._model.fit(
            graph,
            node_counts=node_counts,
            node_frequencies=node_frequencies
        )

    def _normalize_output(
        self,
        edge_node_ids: np.ndarray,
        similarities: np.ndarray,
        return_similarities_dataframe: bool = False,
        return_node_names: bool = False
    ) -> Union[pd.DataFrame, Tuple[np.ndarray, np.ndarray]]:
        """Normalize output to provided standard.
        
        Parameters
        ---------------------
        edge_node_ids: np.ndarray
            The edge node IDs composing the edges.
        similarities: np.ndarray
            Resnik similarity scores.
        return_similarities_dataframe: bool = False
            Whether to return the data as a DataFrame.
            Do note that this will require more RAM.
        return_node_names: bool = False
            Whether to return the node names.
            Do note that this will require SIGNIFICANTLY more RAM.
        """
        if return_node_names and not return_similarities_dataframe:
            raise NotImplementedError(
                "It is not currently supported to return the node names "
                "when it is not requested to return a pandas DataFrame. "
                "This is not supported as the RAM requirements for common "
                "use cases are large enough to be unfeaseable on most systems."
            )

        if not return_similarities_dataframe:
            return (edge_node_ids, similarities)

        return pd.DataFrame({
            "source": (
                self._graph.get_node_names_from_node_ids(edge_node_ids[:, 0])
                if return_node_names
                else edge_node_ids[:, 0]
            ),
            "destination": (
                self._graph.get_node_names_from_node_ids(edge_node_ids[:, 1])
                if return_node_names
                else edge_node_ids[:, 1]
            ),
            "resnik_score": similarities
        })

    def get_similarities_from_bipartite_graph_node_ids(
        self,
        source_node_ids: List[str],
        destination_node_ids: List[str],
        minimum_similarity: Optional[float] = 0.0,
        return_similarities_dataframe: bool = False,
        return_node_names: bool = False
    ) -> Union[pd.DataFrame, Tuple[np.ndarray, np.ndarray]]:
        """Execute similarities probabilities on the bipartite portion

        Parameters
        --------------------
        source_node_ids: List[int]
            The source node ids defining a bipartite graph.
        destination_node_ids: List[int]
            The destination node ids defining a bipartite graph.
        minimum_similarity: Optional[float] = 0.0
            Minimum similarity to be kept. Values below this amount are filtered.
        return_similarities_dataframe: bool = False
            Whether to return a pandas DataFrame, which as indices has the node IDs.
            By default, a numpy array with the similarities is returned as requires much less RAM.
        return_node_names: bool = False
            Whether to return the node names or node IDs associated to the scores.
            By default we return the node ids, which require much less memory.
        """
        if isinstance(source_node_ids, int):
            source_node_ids = [source_node_ids]

        if isinstance(destination_node_ids, int):
            destination_node_ids = [destination_node_ids]
        
        return self._normalize_output(
            *self._model.get_node_ids_and_similarity_from_node_ids(
                first_node_ids=source_node_ids,
                second_node_ids=destination_node_ids,
                minimum_similarity=minimum_similarity
            ),
            return_similarities_dataframe=return_similarities_dataframe,
            return_node_names=return_node_names
        )

    def get_similarities_from_bipartite_graph_node_names(
        self,
        source_node_names: List[str],
        destination_node_names: List[str],
        minimum_similarity: Optional[float] = 0.0,
        return_similarities_dataframe: bool = False,
        return_node_names: bool = False
    ) -> Union[pd.DataFrame, Tuple[np.ndarray, np.ndarray]]:
        """Execute similarities probabilities on the bipartite portion

        Parameters
        --------------------
        source_node_names: List[str]
            The source node names defining a bipartite graph.
        destination_node_names: List[str]
            The destination node names defining a bipartite graph.
        minimum_similarity: Optional[float] = 0.0
            Minimum similarity to be kept. Values below this amount are filtered.
        return_similarities_dataframe: bool = False
            Whether to return a pandas DataFrame, which as indices has the node IDs.
            By default, a numpy array with the similarities is returned as requires much less RAM.
        return_node_names: bool = False
            Whether to return the node names or node IDs associated to the scores.
            By default we return the node ids, which require much less memory.
        """
        if isinstance(source_node_names, str):
            source_node_names = [source_node_names]

        if isinstance(destination_node_names, str):
            destination_node_names = [destination_node_names]
        
        return self._normalize_output(
            *self._model.get_node_ids_and_similarity_from_node_names(
                first_node_names=source_node_names,
                second_node_names=destination_node_names,
                minimum_similarity=minimum_similarity
            ),
            return_similarities_dataframe=return_similarities_dataframe,
            return_node_names=return_node_names
        )

    def get_similarities_from_bipartite_graph_node_prefixes(
        self,
        source_node_prefixes: List[str],
        destination_node_prefixes: List[str],
        minimum_similarity: Optional[float] = 0.0,
        return_similarities_dataframe: bool = False,
        return_node_names: bool = False
    ) -> Union[pd.DataFrame, Tuple[np.ndarray, np.ndarray]]:
        """Execute similarities probabilities on the bipartite portion

        Parameters
        --------------------
        source_node_prefixes: List[str]
            The source node prefixes defining a bipartite graph.
        destination_node_prefixes: List[str]
            The destination node prefixes defining a bipartite graph.
        minimum_similarity: Optional[float] = 0.0
            Minimum similarity to be kept. Values below this amount are filtered.
        return_similarities_dataframe: bool = False
            Whether to return a pandas DataFrame, which as indices has the node IDs.
            By default, a numpy array with the similarities is returned as requires much less RAM.
        return_node_names: bool = False
            Whether to return the node names or node IDs associated to the scores.
            By default we return the node ids, which require much less memory.
        """
        if isinstance(source_node_prefixes, str):
            source_node_prefixes = [source_node_prefixes]

        if isinstance(destination_node_prefixes, str):
            destination_node_prefixes = [destination_node_prefixes]
        
        return self._normalize_output(
            *self._model.get_node_ids_and_similarity_from_node_prefixes(
                first_node_prefixes=source_node_prefixes,
                second_node_prefixes=destination_node_prefixes,
                minimum_similarity=minimum_similarity
            ),
            return_similarities_dataframe=return_similarities_dataframe,
            return_node_names=return_node_names
        )

    def get_similarities_from_bipartite_graph_node_type_ids(
        self,
        source_node_type_ids: List[int],
        destination_node_type_ids: List[int],
        minimum_similarity: Optional[float] = 0.0,
        return_similarities_dataframe: bool = False,
        return_node_names: bool = False
    ) -> Union[pd.DataFrame, Tuple[np.ndarray, np.ndarray]]:
        """Execute similarities probabilities on the bipartite portion

        Parameters
        --------------------
        source_node_type_ids: List[int]
            The source node type ids defining a bipartite graph.
        destination_node_type_ids: List[int]
            The destination node type ids defining a bipartite graph.
        minimum_similarity: Optional[float] = 0.0
            Minimum similarity to be kept. Values below this amount are filtered.
        return_similarities_dataframe: bool = False
            Whether to return a pandas DataFrame, which as indices has the node IDs.
            By default, a numpy array with the similarities is returned as requires much less RAM.
        return_node_names: bool = False
            Whether to return the node names or node IDs associated to the scores.
            By default we return the node ids, which require much less memory.
        """
        if isinstance(source_node_type_ids, int):
            source_node_type_ids = [source_node_type_ids]

        if isinstance(destination_node_type_ids, int):
            destination_node_type_ids = [destination_node_type_ids]
        
        return self._normalize_output(
            *self._model.get_node_ids_and_similarity_from_node_type_ids(
                first_node_type_ids=source_node_type_ids,
                second_node_type_ids=destination_node_type_ids,
                minimum_similarity=minimum_similarity
            ),
            return_similarities_dataframe=return_similarities_dataframe,
            return_node_names=return_node_names
        )

    def get_similarities_from_bipartite_graph_node_type_names(
        self,
        source_node_type_names: List[str],
        destination_node_type_names: List[str],
        minimum_similarity: Optional[float] = 0.0,
        return_similarities_dataframe: bool = False,
        return_node_names: bool = False
    ) -> Union[pd.DataFrame, Tuple[np.ndarray, np.ndarray]]:
        """Execute similarities probabilities on the provided graph bipartite portion.

        Parameters
        --------------------
        source_node_type_names: List[str]
            The source node type names defining a bipartite graph.
        destination_node_type_names: List[str]
            The destination node type names defining a bipartite graph.
        minimum_similarity: Optional[float] = 0.0
            Minimum similarity to be kept. Values below this amount are filtered.
        return_similarities_dataframe: bool = False
            Whether to return a pandas DataFrame, which as indices has the node IDs.
            By default, a numpy array with the similarities is returned as requires much less RAM.
        return_node_names: bool = False
            Whether to return the node names or node IDs associated to the scores.
            By default we return the node ids, which require much less memory.
        """
        if isinstance(source_node_type_names, str):
            source_node_type_names = [source_node_type_names]

        if isinstance(destination_node_type_names, str):
            destination_node_type_names = [destination_node_type_names]

        return self._normalize_output(
            *self._model.get_node_ids_and_similarity_from_node_type_names(
                first_node_type_names=source_node_type_names,
                second_node_type_names=destination_node_type_names,
                minimum_similarity=minimum_similarity
            ),
            return_similarities_dataframe=return_similarities_dataframe,
            return_node_names=return_node_names
        )

    def get_similarities_from_clique_graph_node_ids(
        self,
        node_ids: List[str],
        minimum_similarity: Optional[float] = 0.0,
        return_similarities_dataframe: bool = False,
        return_node_names: bool = False
    ) -> Union[pd.DataFrame, Tuple[np.ndarray, np.ndarray]]:
        """Execute similarities probabilities on the clique portion

        Parameters
        --------------------
        node_ids: List[int]
            The node type ids defining a clique.
        minimum_similarity: Optional[float] = 0.0
            Minimum similarity to be kept. Values below this amount are filtered.
        return_similarities_dataframe: bool = False
            Whether to return a pandas DataFrame, which as indices has the node IDs.
            By default, a numpy array with the similarities is returned as requires much less RAM.
        return_node_names: bool = False
            Whether to return the node names or node IDs associated to the scores.
            By default we return the node ids, which require much less memory.
        """
        return self.get_similarities_from_bipartite_graph_node_ids(
            source_node_ids=node_ids,
            destination_node_ids=node_ids,
            minimum_similarity=minimum_similarity,
            return_similarities_dataframe=return_similarities_dataframe,
            return_node_names=return_node_names
        )

    def get_similarities_from_clique_graph_node_names(
        self,
        node_names: List[str],
        minimum_similarity: Optional[float] = 0.0,
        return_similarities_dataframe: bool = False,
        return_node_names: bool = False
    ) -> Union[pd.DataFrame, Tuple[np.ndarray, np.ndarray]]:
        """Execute similarities probabilities on the clique portion

        Parameters
        --------------------
        node_names: List[str]
            The node type ids defining a clique.
        minimum_similarity: Optional[float] = 0.0
            Minimum similarity to be kept. Values below this amount are filtered.
        return_similarities_dataframe: bool = False
            Whether to return a pandas DataFrame, which as indices has the node IDs.
            By default, a numpy array with the similarities is returned as requires much less RAM.
        return_node_names: bool = False
            Whether to return the node names or node IDs associated to the scores.
            By default we return the node ids, which require much less memory.
        """
        return self.get_similarities_from_bipartite_graph_node_names(
            source_node_names=node_names,
            destination_node_names=node_names,
            minimum_similarity=minimum_similarity,
            return_similarities_dataframe=return_similarities_dataframe,
            return_node_names=return_node_names
        )

    def get_similarities_from_clique_graph_node_prefixes(
        self,
        node_prefixes: List[str],
        minimum_similarity: Optional[float] = 0.0,
        return_similarities_dataframe: bool = False,
        return_node_names: bool = False
    ) -> Union[pd.DataFrame, Tuple[np.ndarray, np.ndarray]]:
        """Execute similarities probabilities on the clique portion

        Parameters
        --------------------
        node_prefixes: List[str]
            The node type ids defining a clique.
        minimum_similarity: Optional[float] = 0.0
            Minimum similarity to be kept. Values below this amount are filtered.
        return_similarities_dataframe: bool = False
            Whether to return a pandas DataFrame, which as indices has the node IDs.
            By default, a numpy array with the similarities is returned as requires much less RAM.
        return_node_names: bool = False
            Whether to return the node names or node IDs associated to the scores.
            By default we return the node ids, which require much less memory.
        """
        return self.get_similarities_from_bipartite_graph_node_prefixes(
            source_node_prefixes=node_prefixes,
            destination_node_prefixes=node_prefixes,
            minimum_similarity=minimum_similarity,
            return_similarities_dataframe=return_similarities_dataframe,
            return_node_names=return_node_names
        )

    def get_similarities_from_clique_graph_node_type_ids(
        self,
        node_type_ids: List[int],
        minimum_similarity: Optional[float] = 0.0,
        return_similarities_dataframe: bool = False,
        return_node_names: bool = False
    ) -> Union[pd.DataFrame, Tuple[np.ndarray, np.ndarray]]:
        """Execute similarities probabilities on the clique portion

        Parameters
        --------------------
        node_type_ids: List[int]
            The node type ids defining a clique.
        minimum_similarity: Optional[float] = 0.0
            Minimum similarity to be kept. Values below this amount are filtered.
        return_similarities_dataframe: bool = False
            Whether to return a pandas DataFrame, which as indices has the node IDs.
            By default, a numpy array with the similarities is returned as requires much less RAM.
        return_node_names: bool = False
            Whether to return the node names or node IDs associated to the scores.
            By default we return the node ids, which require much less memory.
        """
        return self.get_similarities_from_bipartite_graph_node_type_ids(
            source_node_type_ids=node_type_ids,
            destination_node_type_ids=node_type_ids,
            minimum_similarity=minimum_similarity,
            return_similarities_dataframe=return_similarities_dataframe,
            return_node_names=return_node_names
        )

    def get_similarities_from_clique_graph_node_type_names(
        self,
        node_type_names: List[str],
        minimum_similarity: Optional[float] = 0.0,
        return_similarities_dataframe: bool = False,
        return_node_names: bool = False
    ) -> Union[pd.DataFrame, Tuple[np.ndarray, np.ndarray]]:
        """Execute similarities probabilities on the provided graph clique portion.

        Parameters
        --------------------
        node_type_names: List[str]
            The node type names defining a clique.
        minimum_similarity: Optional[float] = 0.0
            Minimum similarity to be kept. Values below this amount are filtered.
        return_similarities_dataframe: bool = False
            Whether to return a pandas DataFrame, which as indices has the node IDs.
            By default, a numpy array with the similarities is returned as requires much less RAM.
        return_node_names: bool = False
            Whether to return the node names or node IDs associated to the scores.
            By default we return the node ids, which require much less memory.
        """
        return self.get_similarities_from_bipartite_graph_node_type_names(
            source_node_type_names=node_type_names,
            destination_node_type_names=node_type_names,
            minimum_similarity=minimum_similarity,
            return_similarities_dataframe=return_similarities_dataframe,
            return_node_names=return_node_names
        )



"""Module providing adapter class making node-label prediction possible in sklearn models."""
from typing import Type, List, Dict, Optional
import numpy as np
import compress_pickle
import copy
from ensmallen import Graph
from embiggen.embedding_transformers import (
    NodeLabelPredictionTransformer,
    NodeTransformer,
)
from embiggen.node_label_prediction.node_label_prediction_model import (
    AbstractNodeLabelPredictionModel,
)
from embiggen.utils.abstract_models import abstract_class


@abstract_class
class SklearnLikeNodeLabelPredictionAdapter(AbstractNodeLabelPredictionModel):
    """Class wrapping Sklearn models for running node-label predictions."""

    def __init__(
        self, model_instance, random_state: Optional[int] = None
    ):
        """Create the adapter for Sklearn object.

        Parameters
        ----------------
        model_instance: Model
            The class instance to be adapted into node-label prediction.
        random_state: Optional[int] = None
            The random state to use to reproduce the training.

        Raises
        ----------------
        ValueError
            If the provided model_instance is not a subclass of `ClassifierMixin`.
        """
        super().__init__(random_state=random_state)
        self._model_instance = model_instance

    def clone(self):
        """Return copy of self."""
        return copy.deepcopy(self)

    def _trasform_graph_into_node_embedding(
        self,
        graph: Graph,
        node_features: List[np.ndarray],
    ) -> np.ndarray:
        """Transforms the provided data into an Sklearn-compatible numpy array.

        Parameters
        ------------------
        graph: Graph,
            The graph whose edges are to be embedded and predicted.
            It can either be an Graph or a list of lists of edges.
        node_features: List[np.ndarray]
            The node features to be used in the training of the model.

        Raises
        ------------------
        ValueError
            If the two graphs do not share the same node vocabulary.
        """
        gt = NodeTransformer(aligned_mapping=True)
        gt.fit(node_features)
        return gt.transform(
            graph,
        )

    def _fit(
        self,
        graph: Graph,
        support: Optional[Graph] = None,
        node_features: Optional[List[np.ndarray]] = None,
        node_type_features: Optional[List[np.ndarray]] = None,
        edge_type_features: Optional[List[np.ndarray]] = None,
        edge_features: Optional[List[np.ndarray]] = None,
    ):
        """Execute fitting of the model.

        Parameters
        ------------------
        graph: Graph,
            The graph whose edges are to be embedded and edge types extracted.
            It can either be an Graph or a list of lists of edges.
        support: Optional[Graph] = None
            The graph describiding the topological structure that
            includes also the above graph. This parameter
            is mostly useful for topological classifiers
            such as Graph Convolutional Networks.
        node_features: Optional[List[np.ndarray]] = None
            The node features to be used in the training of the model.
        node_type_features: Optional[List[np.ndarray]] = None
            The node type features to be used in the training of the model.
        edge_type_features: Optional[List[np.ndarray]] = None
            The edge type features to be used in the training of the model.
        edge_features: Optional[List[np.ndarray]] = None
            Optional edge features to be used as input Concatenated
            to the obtained edge embedding. The shape must be equal
            to the number of directed edges in the graph.

        Raises
        ------------------
        ValueError
            If the two graphs do not share the same node vocabulary.
        """
        nlpt = NodeLabelPredictionTransformer(aligned_mapping=True)

        nlpt.fit(node_features)

        self._model_instance.fit(
            *nlpt.transform(
                graph=graph,
                behaviour_for_unknown_node_labels="drop",
                shuffle=True,
                random_state=self._random_state,
            )
        )

    def _predict_proba(
        self,
        graph: Graph,
        support: Optional[Graph] = None,
        node_features: Optional[List[np.ndarray]] = None,
        node_type_features: Optional[List[np.ndarray]] = None,
        edge_type_features: Optional[List[np.ndarray]] = None,
        edge_features: Optional[List[np.ndarray]] = None,
    ) -> Dict[str, float]:
        """Return evaluations of the model on the edge-label prediction task on the provided data.

        Parameters
        ------------------
        graph: Graph,
            The graph whose edges are to be embedded and predicted.
            It can either be an Graph or a list of lists of edges.
        support: Optional[Graph] = None
            The graph describiding the topological structure that
            includes also the above graph. This parameter
            is mostly useful for topological classifiers
            such as Graph Convolutional Networks.
        node_features: Optional[List[np.ndarray]] = None
            The node features to be used in the evaluation of the model.
        node_type_features: Optional[List[np.ndarray]] = None
            The node type features to be used in prediction.
        edge_type_features: Optional[List[np.ndarray]] = None
            The edge type features to be used in prediction.
        edge_features: Optional[List[np.ndarray]] = None
            Optional edge features to be used as input Concatenated
            to the obtained edge embedding. The shape must be equal
            to the number of directed edges in the provided graph.

        Raises
        ------------------
        ValueError
            If the two graphs do not share the same node vocabulary.
        """
        features = self._trasform_graph_into_node_embedding(
            graph=graph,
            node_features=node_features,
        )

        if hasattr(self._model_instance, "predict_proba"):
            predictions_probabilities = self._model_instance.predict_proba(features)
        elif self.is_multilabel_prediction_task():
            predictions_probabilities = self._model_instance.predict(features)
        else:
            predictions = self._model_instance.predict(features).astype(np.int32)
            predictions_probabilities = np.zeros(
                (predictions.shape[0], len(self._model_instance.classes_)),
                dtype=np.float32,
            )
            predictions_probabilities[np.arange(predictions.size), predictions] = 1

        if self.is_multilabel_prediction_task():
            if isinstance(predictions_probabilities, np.ndarray):
                return predictions_probabilities
            if isinstance(predictions_probabilities, list):
                return np.array(
                    [
                        class_predictions[:, 1]
                        for class_predictions in predictions_probabilities
                    ]
                ).T
            raise NotImplementedError(
                f"The model {self.model_name()} from library {self.library_name()} "
                f"returned an object of type {type(predictions_probabilities)} during "
                "the execution of the predict proba method."
            )

        return predictions_probabilities

    def _predict(
        self,
        graph: Graph,
        support: Optional[Graph] = None,
        node_features: Optional[List[np.ndarray]] = None,
        node_type_features: Optional[List[np.ndarray]] = None,
        edge_type_features: Optional[List[np.ndarray]] = None,
        edge_features: Optional[List[np.ndarray]] = None,
    ) -> Dict[str, float]:
        """Return evaluations of the model on the edge-label prediction task on the provided data.

        Parameters
        ------------------
        graph: Graph,
            The graph whose edges are to be embedded and predicted.
            It can either be an Graph or a list of lists of edges.
        support: Optional[Graph] = None
            The graph describiding the topological structure that
            includes also the above graph. This parameter
            is mostly useful for topological classifiers
            such as Graph Convolutional Networks.
        node_features: List[np.ndarray]
            The node features to be used in prediction.
        node_type_features: List[np.ndarray]
            The node features to be used in prediction.
        edge_type_features: Optional[List[np.ndarray]] = None
            The edge type features to be used in prediction.
        edge_features: Optional[List[np.ndarray]] = None
            Optional edge features to be used as input Concatenated
            to the obtained edge embedding. The shape must be equal
            to the number of directed edges in the provided graph.

        Raises
        ------------------
        ValueError
            If the two graphs do not share the same node vocabulary.
        """
        return self._model_instance.predict(
            self._trasform_graph_into_node_embedding(
                graph=graph,
                node_features=node_features,
            )
        )

    @classmethod
    def can_use_edge_weights(cls) -> bool:
        """Returns whether the model can optionally use edge weights."""
        return False

    @classmethod
    def can_use_edge_types(cls) -> bool:
        """Returns whether the model can optionally use edge types."""
        return False

    @classmethod
    def load(cls, path: str):
        """Load a saved version of the model from the provided path.

        Parameters
        -------------------
        path: str
            Path from where to load the model.
        """
        return compress_pickle.load(path)

    def dump(self, path: str):
        """Dump the current model at the provided path.

        Parameters
        -------------------
        path: str
            Path from where to dump the model.
        """
        compress_pickle.dump(self, path)




"""Unit test class for GraphTransformer objects."""
from platform import node
from typing import Callable
from tqdm.auto import tqdm
from unittest import TestCase
import pytest
import numpy as np
from embiggen.edge_prediction.edge_prediction_model import AbstractEdgePredictionModel
from embiggen.similarities import DAGResnik
from ensmallen.datasets.kgobo import HP


class TestDAGResnik(TestCase):
    """Unit test class for GraphTransformer objects."""

    def setUp(self):
        """Setup objects for running tests on GraphTransformer objects class."""
        self._graph = HP(directed=True)\
            .filter_from_names(
                edge_type_names_to_keep=["biolink:subclass_of"],
                node_prefixes_to_keep=["HP:"]
        )\
            .to_transposed()\
            .remove_disconnected_nodes()
        self._graph.enable()
        self._model = DAGResnik()
        self._model.fit(
            self._graph,
            node_counts={
                node_name: 1
                for node_name in self._graph.get_node_names()
            }
        )
    
    def generic_test(self, callback):
        for return_similarities_dataframe in (True, False):
            for return_node_names in (False, True):
                if return_node_names and not return_similarities_dataframe:
                    with pytest.raises(NotImplementedError):
                        callback(
                        return_similarities_dataframe,
                        return_node_names
                    )
                else:
                    callback(
                        return_similarities_dataframe,
                        return_node_names
                    )
    
    def test_resnik_node_ids(self):
        self.generic_test(
            lambda x, y: self._model.get_similarities_from_clique_graph_node_ids(
                node_ids=[0, 1],
                minimum_similarity=1,
                return_similarities_dataframe=x,
                return_node_names=y
            )
        )

    def test_resnik_node_names(self):
        self.generic_test(
            lambda x, y: self._model.get_similarities_from_clique_graph_node_names(
                node_names=["HP:0000001", "HP:0000002"],
                minimum_similarity=1,
                return_similarities_dataframe=x,
                return_node_names=y
            )
        )
    
    def test_resnik_node_prefixes(self):
        self.generic_test(
            lambda x, y: self._model.get_similarities_from_clique_graph_node_prefixes(
                node_prefixes=["HP:00000"],
                minimum_similarity=1,
                return_similarities_dataframe=x,
                return_node_names=y
            )
        )
    
    def test_resnik_node_type_ids(self):
        self.generic_test(
            lambda x, y: self._model.get_similarities_from_clique_graph_node_type_ids(
                node_type_ids=[0],
                minimum_similarity=1,
                return_similarities_dataframe=x,
                return_node_names=y
            )
        )
    
    def test_resnik_node_type_names(self):
        self.generic_test(
            lambda x, y: self._model.get_similarities_from_clique_graph_node_type_names(
                node_type_names=["biolink:PhenotypicFeature"],
                minimum_similarity=1,
                return_similarities_dataframe=x,
                return_node_names=y
            )
        )



"""Submodule providing edge-label prediction models based on Sklearn Models."""
from embiggen.edge_label_prediction.edge_label_prediction_model import AbstractEdgeLabelPredictionModel
from embiggen.utils.abstract_models import build_init

build_init(
    module_library_names="tensorflow",
    formatted_library_name="TensorFlow",
    task_name="Edge Label Prediction",
    expected_parent_class=AbstractEdgeLabelPredictionModel
)


"""Submodule wrapping Logistic Regression Cross Validator for Edge prediction."""
from typing import Dict, Any, Optional, Union, List
from sklearn.linear_model import LogisticRegressionCV
from embiggen.edge_prediction.edge_prediction_sklearn.sklearn_edge_prediction_adapter import (
    SklearnEdgePredictionAdapter,
)
from embiggen.utils import normalize_kwargs


class LogisticRegressionCVEdgePrediction(SklearnEdgePredictionAdapter):
    """Create wrapper over Sklearn Logistic Regression Cross Validator for Edge prediction."""

    def __init__(
        self,
        Cs: int = 10,
        fit_intercept: bool = True,
        cv: int = 10,
        dual: bool = False,
        penalty: str = "l2",
        scoring: str = "f1_macro",
        solver: str = "lbfgs",
        tol: float = 1e-4,
        max_iter: int = 100,
        class_weight: Optional[str] = "balanced",
        n_jobs: int = -1,
        verbose: int = 0,
        refit: bool = True,
        intercept_scaling: float = 1.0,
        multi_class: str = "auto",
        l1_ratios=None,
        edge_embedding_methods: Union[List[str], str] = "Concatenate",
        training_unbalance_rate: float = 1.0,
        training_sample_only_edges_with_heterogeneous_node_types: bool = False,
        use_edge_metrics: bool = False,
        use_scale_free_distribution: bool = True,
        prediction_batch_size: int = 2**12,
        random_state: int = 42
    ):
        """Create the Logistic Regression Cross Validator for Edge prediction."""
        self._kwargs = normalize_kwargs(
            self,
            dict(
                Cs=Cs,
                fit_intercept=fit_intercept,
                cv=cv,
                dual=dual,
                penalty=penalty,
                scoring=scoring,
                solver=solver,
                tol=tol,
                max_iter=max_iter,
                class_weight=class_weight,
                n_jobs=n_jobs,
                verbose=verbose,
                refit=refit,
                intercept_scaling=intercept_scaling,
                multi_class=multi_class,
                l1_ratios=l1_ratios,
            )
        )

        super().__init__(
            LogisticRegressionCV(**self._kwargs, random_state=random_state),
            edge_embedding_methods=edge_embedding_methods,
            training_unbalance_rate=training_unbalance_rate,
            use_edge_metrics=use_edge_metrics,
            use_scale_free_distribution=use_scale_free_distribution,
            training_sample_only_edges_with_heterogeneous_node_types=training_sample_only_edges_with_heterogeneous_node_types,
            prediction_batch_size=prediction_batch_size,
            random_state=random_state
        )

    def parameters(self) -> Dict[str, Any]:
        """Returns parameters used for this model."""
        return {**super().parameters(), **self._kwargs}

    @classmethod
    def smoke_test_parameters(cls) -> Dict[str, Any]:
        """Returns parameters for smoke test."""
        return dict(cv=2)

    @classmethod
    def model_name(cls) -> str:
        return "Logistic Regression Cross Validator"



"""Module providing an abstract interface for feature preprocessors.

A feature preprocessor is an object that can be used to preprocess the input features
before they are used to fit a model. This can be a very relevant step in the pipeline
of a model execution, and as such, we believe it should be part of the model itself.

This way, when the model is saved, the feature preprocessor is saved as well, and
when the model is loaded, the feature preprocessor is loaded as well.
"""

from typing import Optional, Union, List
import pandas as pd
import numpy as np
import warnings
from ensmallen import Graph
from embiggen.utils.abstract_models.abstract_model import AbstractModel
from embiggen.utils.abstract_models.embedding_result import EmbeddingResult


class AbstractFeaturePreprocessor(AbstractModel):
    """Abstract class for feature preprocessors."""

    def __init__(self, random_state: Optional[int] = None):
        """Create new abstract feature preprocessor.

        Parameters
        -------------------------
        random_state: Optional[int] = None
            The random state to use if the model is stocastic.
        """
        super().__init__(random_state=random_state)

    @classmethod
    def task_name(cls) -> str:
        return "Feature Preprocessor"

    def _transform(
        self,
        support: Graph,
        node_features: List[Union[pd.DataFrame, np.ndarray]],
    ) -> EmbeddingResult:
        """Transform the given node features.

        Parameters
        -------------------------
        node_features: List[Union[pd.DataFrame, np.ndarray]],
            Node feature to use to fit the transformer.
        support: Graph
            Support graph to use to transform the node features.
        """
        raise NotImplementedError(
            "The method _transform should be implemented by the child class."
        )

    def transform(
        self,
        support: Graph,
        node_features: Union[pd.DataFrame, np.ndarray, List[Union[pd.DataFrame, np.ndarray]]],
    ) -> EmbeddingResult:
        """Transform the given node features.

        Parameters
        -------------------------
        support: Graph
            Support graph to use to transform the node features.
        node_features: Union[pd.DataFrame, np.ndarray, List[Union[pd.DataFrame, np.ndarray]]],
            Node feature to use to fit the transformer.
        """
        assert isinstance(support, Graph)

        if not isinstance(node_features, list):
            node_features = [node_features]

        if len(node_features) == 0:
            raise ValueError(
                "You have provided an empty list of node features "
                f"to the {self.model_name()} feature preprocessor."
            )

        if len(node_features) > 1:
            warnings.warn(
                "You have provided more than one node feature to the "
                f"{self.model_name()} feature preprocessor. "
                "This feature preprocessor will be applied to each of them "
                "independently."
            )

        for feature in node_features:
            assert isinstance(feature, (pd.DataFrame, np.ndarray))

        return self._transform(support=support, node_features=node_features)



"""Wrapper for MNMF model provided from the Karate Club package."""
from typing import Dict, Any
from karateclub.community_detection import MNMF
from embiggen.embedders.karateclub_embedders.abstract_karateclub_embedder import AbstractKarateClubEmbedder


class MNMFKarateClub(AbstractKarateClubEmbedder):

    def __init__(
        self,
        embedding_size: int = 128,
        clusters: int = 10,
        lambd: float = 0.2,
        alpha: float = 0.05,
        beta: float = 0.05,
        iterations: int = 200,
        lower_control: float = 10**-15,
        eta: float = 5.0,
        random_state: int = 42,
        ring_bell: bool = False,
        enable_cache: bool = False
    ):
        """Return a new MNMF embedding model.

        Parameters
        ----------------------
        embedding_size: int = 128
            Number of dimensions. Default is 128.
        clusters: int = 10
            Number of clusters. Default is 10.
        lambd: float = 0.2
            KKT penalty. Default is 0.2
        alpha: float = 0.05
            Clustering penalty. Default is 0.05.
        beta: float = 0.05
            Modularity regularization penalty. Default is 0.05.
        iterations: int = 200
            Number of power iterations. Default is 200.
        lower_control: float = 10**-15
            Floating point overflow control. Default is 10**-15.
        eta: float = 5.0
            Similarity mixing parameter. Default is 5.0.
        random_state: int = 42
            Random state to use for the stocastic
            portions of the embedding algorithm.
        ring_bell: bool = False,
            Whether to play a sound when embedding completes.
        enable_cache: bool = False
            Whether to enable the cache, that is to
            store the computed embedding.
        """
        self._clusters = clusters
        self._lambd = lambd
        self._alpha = alpha
        self._beta = beta
        self._iterations = iterations
        self._lower_control = lower_control
        self._eta = eta

        super().__init__(
            embedding_size=embedding_size,
            enable_cache=enable_cache,
            ring_bell=ring_bell,
            random_state=random_state
        )

    def parameters(self) -> Dict[str, Any]:
        """Returns the parameters used in the model."""
        return dict(
            **super().parameters(),
            clusters = self._clusters,
            lambd=self._lambd,
            alpha=self._alpha,
            beta=self._beta,
            iterations=self._iterations,
            lower_control=self._lower_control,
            eta=self._eta,
        )

    @classmethod
    def smoke_test_parameters(cls) -> Dict[str, Any]:
        """Returns parameters for smoke test."""
        return dict(
            **AbstractKarateClubEmbedder.smoke_test_parameters(),
            iterations=1,
        )

    def _build_model(self) -> MNMF:
        """Return new instance of the MNMF model."""
        return MNMF(
            dimensions=self._embedding_size,
            clusters = self._clusters,
            lambd=self._lambd,
            alpha=self._alpha,
            beta=self._beta,
            iterations=self._iterations,
            lower_control=self._lower_control,
            eta=self._eta,
            seed=self._random_state
        )

    @classmethod
    def model_name(cls) -> str:
        """Returns name of the model"""
        return "MNMF"

    def get_cluster_centers(self):
        return self._model.get_cluster_centers()

    def get_memberships(self):
        """Getting the cluster membership of nodes."""
        return self._model.get_memberships()

    @classmethod
    def requires_nodes_sorted_by_decreasing_node_degree(cls) -> bool:
        return False

    @classmethod
    def is_topological(cls) -> bool:
        return True

    @classmethod
    def can_use_edge_weights(cls) -> bool:
        """Returns whether the model can optionally use edge weights."""
        return False

    @classmethod
    def can_use_node_types(cls) -> bool:
        """Returns whether the model can optionally use node types."""
        return False

    @classmethod
    def can_use_edge_types(cls) -> bool:
        """Returns whether the model can optionally use edge types."""
        return False

